#!/usr/bin/env python3
# -*- coding:utf-8 -*-
"""
@file    python_convolution_performance_tests.py

@author  Lars Pastewka <lars.pastewka@imtek.uni-freiburg.de>

@date    12 Nov 2025

@brief   Performance tests for convolution operators

Copyright © 2025 Lars Pastewka

µGrid is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3, or (at
your option) any later version.

µGrid is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with µGrid; see the file COPYING. If not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.

Additional permission under GNU GPL version 3 section 7

If you modify this Program, or any covered work, by linking or combining it
with proprietary FFT implementations or numerical libraries, containing parts
covered by the terms of those libraries' licenses, the licensors of this
Program grant you additional permission to convey the resulting work.
"""

import time
import unittest
from dataclasses import dataclass
from typing import Optional

import numpy as np
from scipy.sparse import coo_array

import muGrid
from muGrid import real_field, wrap_field
from muGrid.Solvers import conjugate_gradients

# Try to import pypapi for hardware counter access
try:
    import pypapi
    from pypapi import events as papi_events

    PAPI_AVAILABLE = True
except ImportError:
    PAPI_AVAILABLE = False
    print("Warning: pypapi not available. Install with: pip install pypapi")
    print("Performance tests will run without hardware counters.")

# Try to import CuPy for GPU tests
try:
    import cupy as cp

    HAS_CUPY = True
except ImportError:
    HAS_CUPY = False


# Use compile-time feature flag for GPU availability
GPU_AVAILABLE = muGrid.has_gpu


def get_gpu_skip_reason():
    """Get detailed skip reason for GPU tests."""
    if not GPU_AVAILABLE:
        return "GPU backend not available (muGrid not compiled with CUDA/HIP)"
    return None


def get_gpu_cupy_skip_reason():
    """Get detailed skip reason for GPU+CuPy tests."""
    reasons = []
    if not GPU_AVAILABLE:
        reasons.append("GPU backend not available (muGrid not compiled with CUDA/HIP)")
    if not HAS_CUPY:
        reasons.append("CuPy not installed (pip install cupy-cuda*)")
    if reasons:
        return "; ".join(reasons)
    return None


@dataclass
class PerformanceMetrics:
    """Container for performance metrics"""

    wall_time_ms: float
    total_flops: int
    grid_size: tuple
    stencil_size: tuple
    nb_components: int
    nb_operators: int
    nb_quad_pts: int

    # Device info
    device: str = "cpu"

    # Hardware counters (optional)
    papi_cycles: Optional[int] = None
    papi_instructions: Optional[int] = None
    papi_fp_ops: Optional[int] = None
    papi_l1_dcm: Optional[int] = None
    papi_l2_dcm: Optional[int] = None
    papi_l3_tcm: Optional[int] = None

    @property
    def gflops(self) -> float:
        """Calculate GFLOPS from theoretical FLOPs"""
        return (self.total_flops / 1e9) / (self.wall_time_ms / 1000.0)

    @property
    def actual_gflops(self) -> float:
        """Calculate GFLOPS from PAPI measured FP operations"""
        if self.papi_fp_ops is not None:
            return (self.papi_fp_ops / 1e9) / (self.wall_time_ms / 1000.0)
        return self.gflops

    @property
    def ipc(self) -> Optional[float]:
        """Calculate instructions per cycle"""
        if self.papi_cycles is not None and self.papi_instructions is not None:
            return self.papi_instructions / self.papi_cycles
        return None

    @property
    def l1_miss_rate(self) -> Optional[float]:
        """Calculate L1 cache miss rate as percentage"""
        if self.papi_l1_dcm is not None and self.papi_instructions is not None:
            return (self.papi_l1_dcm / self.papi_instructions) * 100.0
        return None

    @property
    def l2_miss_rate(self) -> Optional[float]:
        """Calculate L2 cache miss rate as percentage"""
        if self.papi_l2_dcm is not None and self.papi_instructions is not None:
            return (self.papi_l2_dcm / self.papi_instructions) * 100.0
        return None

    @property
    def l3_miss_rate(self) -> Optional[float]:
        """Calculate L3 cache miss rate as percentage"""
        if self.papi_l3_tcm is not None and self.papi_instructions is not None:
            return (self.papi_l3_tcm / self.papi_instructions) * 100.0
        return None

    def print_report(self):
        """Print a formatted performance report"""
        print("\n" + "=" * 70)
        print("PERFORMANCE METRICS")
        print("=" * 70)
        print(f"Device:             {self.device}")
        print(
            f"Grid size:          {self.grid_size[0]} x {self.grid_size[1]} "
            f"= {np.prod(self.grid_size)} pixels"
        )
        print(f"Stencil size:       {self.stencil_size[0]} x {self.stencil_size[1]}")
        print(f"Field components:   {self.nb_components}")
        print(f"Operators:          {self.nb_operators}")
        print(f"Quadrature points:  {self.nb_quad_pts}")
        print("-" * 70)
        print(f"Wall time:          {self.wall_time_ms:.3f} ms")
        print(f"Theoretical FLOPs:  {self.total_flops:,}")
        print(f"Theoretical GFLOPS: {self.gflops:.3f}")

        if self.papi_fp_ops is not None:
            print("-" * 70)
            print("PAPI Hardware Counters:")
            print(f"  Actual FP Ops:    {self.papi_fp_ops:,}")
            print(f"  Actual GFLOPS:    {self.actual_gflops:.3f}")
            print(f"  Total cycles:     {self.papi_cycles:,}")
            print(f"  Total instr:      {self.papi_instructions:,}")
            print(f"  IPC:              {self.ipc:.2f}")
            print(
                f"  L1 cache misses:  {self.papi_l1_dcm:,} "
                f"({self.l1_miss_rate:.2f}% of instructions)"
            )
            print(
                f"  L2 cache misses:  {self.papi_l2_dcm:,} "
                f"({self.l2_miss_rate:.2f}% of instructions)"
            )
            print(
                f"  L3 cache misses:  {self.papi_l3_tcm:,} "
                f"({self.l3_miss_rate:.2f}% of instructions)"
            )
        print("=" * 70 + "\n")


def count_theoretical_flops(conv_op, nodal_field, quad_field, grid_size):
    """
    Count theoretical FLOPs for a convolution operation.

    The convolution operation at each pixel involves:
    - Looping over stencil points (nb_conv_pts)
    - For each stencil point, looping over nodal sub-points (nb_pixelnodal_pts)
    - Computing: quad_vals += alpha * nodal_vals * operator_vals

    Each outer product (vector * vector) involves:
    - nb_components multiplications (for the column vector)
    - nb_operators * nb_quad_pts multiplications (for broadcasting)
    - Same number of additions

    Total per pixel: 2 * nb_components * nb_operators * nb_quad_pts *
        nb_conv_pts * nb_nodal_pts
    """
    nb_pixels = np.prod(grid_size)
    nb_components = nodal_field.nb_components
    nb_operators = conv_op.nb_operators
    nb_quad_pts = conv_op.nb_quad_pts
    nb_nodal_pts = conv_op.nb_nodal_pts

    # Get stencil shape from the operator
    # This is a bit tricky - we need to compute nb_conv_pts from the stencil
    # For now, estimate from the operator matrix dimensions
    # pixel_operator returns a flat list, so we compute nb_conv_pts from total size
    operator_list = conv_op.pixel_operator
    total_elements = len(operator_list)
    # Total elements = nb_operators * nb_quad_pts * nb_nodal_pts * nb_conv_pts
    nb_conv_pts = total_elements // (nb_operators * nb_quad_pts * nb_nodal_pts)

    # Each convolution point contributes:
    # - nb_components values (from nodal field)
    # - multiplied by nb_operators * nb_quad_pts values (from operator)
    # - 2 FLOPs per element (multiply + add)
    flops_per_pixel = (
        2 * nb_components * nb_operators * nb_quad_pts * nb_conv_pts * nb_nodal_pts
    )

    total_flops = flops_per_pixel * nb_pixels

    return total_flops


def measure_convolution_performance(
    conv_op,
    nodal_field,
    quad_field,
    grid_size,
    stencil_size,
    num_iterations=10,
    use_papi=True,
    device="cpu",
):
    """
    Measure performance of convolution operator.

    Parameters
    ----------
    conv_op : muGrid.ConvolutionOperator
        The convolution operator to benchmark
    nodal_field : muGrid.Field
        Input nodal field
    quad_field : muGrid.Field
        Output quadrature field
    grid_size : tuple
        Size of the grid (nx, ny)
    stencil_size : tuple
        Size of the stencil (sx, sy)
    num_iterations : int
        Number of iterations for averaging
    use_papi : bool
        Whether to use PAPI hardware counters
    device : str
        Device string (e.g., "cpu", "cuda:0", "rocm:0")

    Returns
    -------
    PerformanceMetrics
        Container with all performance metrics
    """
    is_on_device = nodal_field.is_on_gpu

    # Count theoretical FLOPs
    total_flops = count_theoretical_flops(conv_op, nodal_field, quad_field, grid_size)

    # Warm-up iteration
    conv_op.apply(nodal_field, quad_field)

    # Synchronize if on device to ensure warm-up is complete
    if is_on_device and HAS_CUPY:
        cp.cuda.Device().synchronize()

    # Initialize PAPI if available, requested, and on host (PAPI doesn't work for GPU)
    papi_values = {}
    if PAPI_AVAILABLE and use_papi and not is_on_device:
        try:
            # Define events to monitor
            papi_events_list = [
                papi_events.PAPI_TOT_CYC,  # Total cycles
                papi_events.PAPI_TOT_INS,  # Total instructions
                papi_events.PAPI_FP_OPS,  # Floating point operations
                papi_events.PAPI_L1_DCM,  # L1 data cache misses
                papi_events.PAPI_L2_DCM,  # L2 data cache misses
                papi_events.PAPI_L3_TCM,  # L3 total cache misses
            ]

            pypapi.papi_high.start_counters(papi_events_list)
            papi_enabled = True
        except Exception as e:
            print(f"Warning: Could not initialize PAPI: {e}")
            papi_enabled = False
    else:
        papi_enabled = False

    # Time the operation
    if is_on_device and HAS_CUPY:
        # Use CUDA events for accurate GPU timing
        start_event = cp.cuda.Event()
        end_event = cp.cuda.Event()

        start_event.record()
        for _ in range(num_iterations):
            conv_op.apply(nodal_field, quad_field)
        end_event.record()
        end_event.synchronize()

        # Get elapsed time in milliseconds
        elapsed_ms = cp.cuda.get_elapsed_time(start_event, end_event)
        wall_time_ms = elapsed_ms / num_iterations
    else:
        # Use wall-clock time for CPU
        start_time = time.perf_counter()

        for _ in range(num_iterations):
            conv_op.apply(nodal_field, quad_field)

        end_time = time.perf_counter()
        wall_time_ms = ((end_time - start_time) / num_iterations) * 1000.0

    # Read PAPI counters
    if papi_enabled:
        try:
            counters = pypapi.papi_high.stop_counters()
            papi_values = {
                "cycles": counters[0] // num_iterations,
                "instructions": counters[1] // num_iterations,
                "fp_ops": counters[2] // num_iterations,
                "l1_dcm": counters[3] // num_iterations,
                "l2_dcm": counters[4] // num_iterations,
                "l3_tcm": counters[5] // num_iterations,
            }
        except Exception as e:
            print(f"Warning: Could not read PAPI counters: {e}")
            papi_values = {}

    # Create metrics object
    metrics = PerformanceMetrics(
        wall_time_ms=wall_time_ms,
        total_flops=total_flops,
        grid_size=grid_size,
        stencil_size=stencil_size,
        nb_components=nodal_field.nb_components,
        nb_operators=conv_op.nb_operators,
        nb_quad_pts=conv_op.nb_quad_pts,
        device=device,
        papi_cycles=papi_values.get("cycles"),
        papi_instructions=papi_values.get("instructions"),
        papi_fp_ops=papi_values.get("fp_ops"),
        papi_l1_dcm=papi_values.get("l1_dcm"),
        papi_l2_dcm=papi_values.get("l2_dcm"),
        papi_l3_tcm=papi_values.get("l3_tcm"),
    )

    return metrics


class ConvolutionPerformanceTests(unittest.TestCase):
    """Test suite for convolution operator performance"""

    def test_performance_small_grid(self):
        """Benchmark convolution on a small 32x32 grid"""
        print("\n=== Testing 2D Convolution Performance (Small Grid) ===")

        # Parameters
        nb_x_pts = 32
        nb_y_pts = 32
        nb_stencil_x = 3
        nb_stencil_y = 3
        nb_operators = 2
        nb_quad_pts = 4
        nb_field_components = 3

        # Create stencil
        stencil = np.random.rand(nb_operators, nb_quad_pts, nb_stencil_x, nb_stencil_y)

        conv_op = muGrid.ConvolutionOperator([-1, -1], stencil)

        # Create field collection
        nb_ghosts = (1, 1)
        fc = muGrid.GlobalFieldCollection(
            (nb_x_pts, nb_y_pts),
            sub_pts={"quad": nb_quad_pts},
            nb_ghosts_left=nb_ghosts,
            nb_ghosts_right=nb_ghosts,
        )

        # Create fields
        nodal_cpp = fc.real_field("nodal", nb_field_components)
        quad_cpp = fc.real_field("quad", (nb_field_components, nb_operators), "quad")
        nodal = wrap_field(nodal_cpp)

        # Initialize with random data
        nodal.p[...] = np.random.rand(*nodal.p.shape)

        # Measure performance
        metrics = measure_convolution_performance(
            conv_op,
            nodal_cpp,
            quad_cpp,
            grid_size=(nb_x_pts, nb_y_pts),
            stencil_size=(nb_stencil_x, nb_stencil_y),
            num_iterations=100,
        )

        metrics.print_report()

        # Sanity checks
        self.assertGreater(metrics.total_flops, 0)
        self.assertGreater(metrics.gflops, 0.0)

    def test_performance_large_grid(self):
        """Benchmark convolution on a large 256x256 grid"""
        print("\n=== Testing 2D Convolution Performance (Large Grid) ===")

        # Parameters
        nb_x_pts = 256
        nb_y_pts = 256
        nb_stencil_x = 3
        nb_stencil_y = 3
        nb_operators = 2
        nb_quad_pts = 4
        nb_field_components = 3

        # Create stencil
        stencil = np.random.rand(
            nb_operators, nb_quad_pts, 1, nb_stencil_x, nb_stencil_y
        )

        conv_op = muGrid.ConvolutionOperator([-1, -1], stencil)

        # Create field collection
        comm = muGrid.Communicator()
        subdivisions = (1, 1)
        decomposition = muGrid.CartesianDecomposition(
            comm, (nb_x_pts, nb_y_pts), subdivisions, (1, 1), (1, 1)
        )
        fc = decomposition.collection
        fc.set_nb_sub_pts("quad", nb_quad_pts)

        # Create fields
        nodal_cpp = fc.real_field("nodal", nb_field_components)
        quad_cpp = fc.real_field("quad", (nb_field_components, nb_operators), "quad")
        nodal = wrap_field(nodal_cpp)

        # Initialize with random data
        nodal.p[...] = np.random.rand(*nodal.p.shape)

        # Measure performance (fewer iterations for large grid)
        metrics = measure_convolution_performance(
            conv_op,
            nodal_cpp,
            quad_cpp,
            grid_size=(nb_x_pts, nb_y_pts),
            stencil_size=(nb_stencil_x, nb_stencil_y),
            num_iterations=10,
        )

        metrics.print_report()

        # Sanity checks
        self.assertGreater(metrics.total_flops, 0)
        self.assertGreater(metrics.gflops, 0.0)

        if metrics.l1_miss_rate is not None:
            # Cache miss rate should be reasonable
            self.assertLess(metrics.l1_miss_rate, 50.0)

    def test_performance_comparison_grid_sizes(self):
        """Compare performance across different grid sizes"""
        print("\n=== Comparing Performance Across Grid Sizes ===")

        grid_sizes = [16, 32, 64, 128, 256]
        results = []

        for grid_size in grid_sizes:
            nb_stencil = 3
            nb_operators = 2
            nb_quad_pts = 4
            nb_components = 3

            # Create stencil
            stencil = np.random.rand(nb_operators, nb_quad_pts, nb_stencil, nb_stencil)

            conv_op = muGrid.ConvolutionOperator([-1, -1], stencil)

            # Create field collection
            nb_ghosts = (1, 1)
            fc = muGrid.GlobalFieldCollection(
                (grid_size, grid_size),
                sub_pts={"quad": nb_quad_pts},
                nb_ghosts_left=nb_ghosts,
                nb_ghosts_right=nb_ghosts,
            )

            # Create fields
            nodal_cpp = fc.real_field("nodal", nb_components)
            quad_cpp = fc.real_field("quad", (nb_components, nb_operators), "quad")
            nodal = wrap_field(nodal_cpp)

            # Initialize with random data
            nodal.p[...] = np.random.rand(*nodal.p.shape)

            # Measure performance
            iterations = 100 if grid_size <= 64 else 10
            metrics = measure_convolution_performance(
                conv_op,
                nodal_cpp,
                quad_cpp,
                grid_size=(grid_size, grid_size),
                stencil_size=(nb_stencil, nb_stencil),
                num_iterations=iterations,
            )

            results.append(metrics)

        # Print comparison table
        print("\n" + "=" * 90)
        print(
            f"{'Grid Size':>10} {'GFLOPS':>12} {'L1 Miss %':>12} "
            f"{'L2 Miss %':>12} {'L3 Miss %':>12} {'IPC':>12}"
        )
        print("-" * 90)

        for metrics in results:
            grid_pixels = np.prod(metrics.grid_size)
            l1_str = f"{metrics.l1_miss_rate:.2f}" if metrics.l1_miss_rate else "N/A"
            l2_str = f"{metrics.l2_miss_rate:.2f}" if metrics.l2_miss_rate else "N/A"
            l3_str = f"{metrics.l3_miss_rate:.2f}" if metrics.l3_miss_rate else "N/A"
            ipc_str = f"{metrics.ipc:.2f}" if metrics.ipc else "N/A"

            print(
                f"{grid_pixels:>10} {metrics.gflops:>12.3f} {l1_str:>12} "
                f"{l2_str:>12} {l3_str:>12} {ipc_str:>12}"
            )

        print("=" * 90 + "\n")


def run_laplace_solver(nb_grid_pts, use_device=False):
    """
    Run Laplace solver using muGrid convolution operator.

    Parameters
    ----------
    nb_grid_pts : tuple
        Grid size (nx, ny)
    use_device : bool
        If True, use GPU memory; otherwise use CPU

    Returns
    -------
    tuple
        (solution array, elapsed time in seconds, device string)
    """
    comm = muGrid.Communicator()
    subdivisions = (1, 1)

    # Determine memory location
    if use_device:
        memory_location = muGrid.GlobalFieldCollection.MemoryLocation.Device
    else:
        memory_location = muGrid.GlobalFieldCollection.MemoryLocation.Host

    # Setup problem
    decomposition = muGrid.CartesianDecomposition(
        comm, nb_grid_pts, subdivisions, (1, 1), (1, 1), memory_location=memory_location
    )
    fc = decomposition.collection
    grid_spacing = 1 / np.array(nb_grid_pts)

    x, y = decomposition.coords
    i, j = decomposition.icoords

    rhs = real_field(decomposition, "rhs")
    solution = real_field(decomposition, "solution")

    # Initialize RHS - need to handle device arrays differently
    if use_device and HAS_CUPY:
        x_cp = cp.asarray(x)
        y_cp = cp.asarray(y)
        rhs_data = (1 + cp.cos(2 * cp.pi * x_cp) * cp.cos(2 * cp.pi * y_cp)) ** 10
        rhs_data -= cp.mean(rhs_data)
        rhs.p[...] = rhs_data
        solution.s[...] = 0
    else:
        rhs.p[...] = (1 + np.cos(2 * np.pi * x) * np.cos(2 * np.pi * y)) ** 10
        rhs.p[...] -= np.mean(rhs.p)
        solution.s[...] = 0

    # Create Laplace operator
    stencil = np.array([[0, 1, 0], [1, -4, 1], [0, 1, 0]])
    laplace = muGrid.ConvolutionOperator([-1, -1], stencil)

    # Get device string for reporting
    device_str = rhs._cpp.device

    # Define Hessian-vector product
    scale_factor = -np.mean(grid_spacing) ** 2

    def hessp_mugrid(x_field, Ax_field):
        decomposition.communicate_ghosts(x_field)
        laplace.apply(x_field, Ax_field)
        Ax = wrap_field(Ax_field)
        Ax.s[...] /= scale_factor
        return Ax_field

    # Time the solver
    if use_device and HAS_CUPY:
        cp.cuda.Device().synchronize()
        start_event = cp.cuda.Event()
        end_event = cp.cuda.Event()
        start_event.record()

    t_start = time.perf_counter()
    conjugate_gradients(
        comm,
        fc,
        hessp_mugrid,
        rhs._cpp,
        solution._cpp,
        tol=1e-6,
        maxiter=1000,
    )
    t_elapsed = time.perf_counter() - t_start

    if use_device and HAS_CUPY:
        end_event.record()
        end_event.synchronize()
        t_elapsed = cp.cuda.get_elapsed_time(start_event, end_event) / 1000.0

    # Get solution back to host for comparison
    if use_device and HAS_CUPY:
        solution_arr = cp.asnumpy(solution.s)
    else:
        solution_arr = solution.s.copy()

    return solution_arr, t_elapsed, device_str


def test_laplace_mugrid_vs_scipy(nb_grid_pts=(512, 512)):
    """Test Laplace solver: muGrid convolution vs SciPy sparse matrix."""
    comm = muGrid.Communicator()
    subdivisions = (1, 1)

    # Run muGrid solution on host
    mugrid_solution, t_mugrid, device_str = run_laplace_solver(
        nb_grid_pts, use_device=False
    )

    # Setup for scipy comparison
    decomposition = muGrid.CartesianDecomposition(
        comm, nb_grid_pts, subdivisions, (1, 1), (1, 1)
    )
    fc = decomposition.collection
    grid_spacing = 1 / np.array(nb_grid_pts)
    x, y = decomposition.coords
    i, j = decomposition.icoords

    solution = real_field(decomposition, "solution")
    rhs = real_field(decomposition, "rhs")
    rhs.p[...] = (1 + np.cos(2 * np.pi * x) * np.cos(2 * np.pi * y)) ** 10
    rhs.p[...] -= np.mean(rhs.p)
    solution.s[...] = 0

    # scipy Laplace sparse matrix
    nb = np.prod(nb_grid_pts)

    def grid_index(i, j):
        return (i % nb_grid_pts[0]).reshape(-1) * nb_grid_pts[1] + (
            j % nb_grid_pts[1]
        ).reshape(-1)

    laplace_sparse = (
        coo_array(
            (-4 * np.ones(nb), (grid_index(i, j), grid_index(i, j))), shape=(nb, nb)
        )
        + coo_array(
            (np.ones(nb), (grid_index(i, j), grid_index(i + 1, j))), shape=(nb, nb)
        )
        + coo_array(
            (np.ones(nb), (grid_index(i, j), grid_index(i - 1, j))), shape=(nb, nb)
        )
        + coo_array(
            (np.ones(nb), (grid_index(i, j), grid_index(i, j + 1))), shape=(nb, nb)
        )
        + coo_array(
            (np.ones(nb), (grid_index(i, j), grid_index(i, j - 1))), shape=(nb, nb)
        )
    ).tocsr()

    def hessp_scipy(x_field, Ax_field):
        x_arr = wrap_field(x_field)
        Ax = wrap_field(Ax_field)
        Ax.p[...] = (laplace_sparse @ x_arr.p.reshape(-1)).reshape(nb_grid_pts)
        Ax.s[...] /= -np.mean(grid_spacing) ** 2
        return Ax_field

    t_scipy = -time.perf_counter()
    conjugate_gradients(
        comm,
        fc,
        hessp_scipy,
        rhs._cpp,
        solution._cpp,
        tol=1e-6,
        maxiter=1000,
    )
    t_scipy += time.perf_counter()

    # Check solutions agree
    np.testing.assert_allclose(solution.s, mugrid_solution)

    # Print timing result
    print(f"muGrid operator time ({device_str}):  {t_mugrid:.6f} s")
    print(f"SciPy Sparse time:  {t_scipy:.6f} s")

    # Check that the speed is at least comparable
    # assert t_mugrid < 1.05 * t_scipy, (
    #     f"muGrid slower than SciPy sparse: "
    #     f"muGrid {t_mugrid:.6f}s vs. SciPy sparse {t_scipy:.6f}s"
    # )


@unittest.skipUnless(GPU_AVAILABLE and HAS_CUPY, get_gpu_cupy_skip_reason() or "")
def test_laplace_device_vs_host():
    """Compare Laplace solver performance between CPU and GPU.

    This test uses CUDA-aware MPI for ghost communication on device fields.
    """
    nb_grid_pts = (256, 256)

    # Run on host
    host_solution, t_host, host_device = run_laplace_solver(
        nb_grid_pts, use_device=False
    )

    # Run on device
    device_solution, t_device, device_str = run_laplace_solver(
        nb_grid_pts, use_device=True
    )

    # Check solutions match
    # Note: CPU and GPU may have small differences due to:
    # - Different floating-point operation ordering
    # - FMA (fused multiply-add) vs separate mul+add
    # - Parallel reduction accumulation order
    np.testing.assert_allclose(
        device_solution,
        host_solution,
        rtol=1e-4,
        atol=1e-5,
        err_msg="Device and host Laplace solutions differ",
    )

    # Print timing comparison
    speedup = t_host / t_device if t_device > 0 else 0
    print(f"\nLaplace solver comparison ({nb_grid_pts[0]}x{nb_grid_pts[1]}):")
    print(f"  Host ({host_device}):   {t_host:.6f} s")
    print(f"  Device ({device_str}): {t_device:.6f} s")
    print(f"  Speedup: {speedup:.2f}x")


@unittest.skipUnless(GPU_AVAILABLE and HAS_CUPY, get_gpu_cupy_skip_reason() or "")
def test_laplace_device_scaling():
    """Test Laplace solver scaling on GPU across different grid sizes.

    This test uses CUDA-aware MPI for ghost communication on device fields.
    """
    grid_sizes = [64, 128, 256, 512]
    results = []

    print("\n=== Laplace Solver GPU Scaling ===")
    print(f"{'Grid':>10} {'Host (s)':>12} {'Device (s)':>12} {'Speedup':>10}")
    print("-" * 50)

    for size in grid_sizes:
        nb_grid_pts = (size, size)

        # Host timing
        _, t_host, _ = run_laplace_solver(nb_grid_pts, use_device=False)

        # Device timing
        _, t_device, device_str = run_laplace_solver(nb_grid_pts, use_device=True)

        speedup = t_host / t_device if t_device > 0 else 0
        results.append((size, t_host, t_device, speedup))

        print(f"{size}x{size:>4} {t_host:>12.6f} {t_device:>12.6f} {speedup:>10.2f}x")

    print("-" * 50)
    print(f"Device: {device_str}")


def get_quad_triangle_kernel():
    """Get the kernel for quadrature on triangles with 6 quadrature points."""
    return np.array(
        [
            [  # operator 1
                [[[2 / 3, 1 / 6], [1 / 6, 0]]],  # quadrature point 1
                [[[1 / 6, 1 / 6], [2 / 3, 0]]],  # quadrature point 2
                [[[1 / 6, 2 / 3], [1 / 6, 0]]],  # quadrature point 3
                [[[0, 1 / 6], [1 / 6, 2 / 3]]],  # quadrature point 4
                [[[0, 2 / 3], [1 / 6, 1 / 6]]],  # quadrature point 5
                [[[0, 1 / 6], [2 / 3, 1 / 6]]],  # quadrature point 6
            ]
        ]
    )


def get_quad_triangle_manual_matrix():
    """Get the manual matrix for quadrature on triangles."""
    return np.array(
        [
            [2 / 3, 1 / 6, 1 / 6, 0],  # q1
            [1 / 6, 2 / 3, 1 / 6, 0],  # q2
            [1 / 6, 1 / 6, 2 / 3, 0],  # q3
            [0, 1 / 6, 1 / 6, 2 / 3],  # q4
            [0, 1 / 6, 2 / 3, 1 / 6],  # q5
            [0, 2 / 3, 1 / 6, 1 / 6],  # q6
        ]
    )


def run_quad_triangle_mugrid(nb_grid_pts, use_device=False):
    """
    Run quadrature triangle computation using muGrid convolution.

    Parameters
    ----------
    nb_grid_pts : tuple
        Grid size (Nx, Ny)
    use_device : bool
        If True, use GPU memory

    Returns
    -------
    tuple
        (result array on host, elapsed time, device string)
    """
    Nx, Ny = nb_grid_pts

    # Determine memory location
    if use_device:
        memory_location = muGrid.GlobalFieldCollection.MemoryLocation.Device
    else:
        memory_location = muGrid.GlobalFieldCollection.MemoryLocation.Host

    # Create field collection
    fc = muGrid.GlobalFieldCollection(
        nb_grid_pts,
        sub_pts={"quad": 6},
        nb_ghosts_right=(1, 1),
        memory_location=memory_location,
    )
    nodal_field_cpp = fc.real_field("nodal")
    quad_field_cpp = fc.real_field("quad", 1, "quad")
    nodal_field = wrap_field(nodal_field_cpp)

    # Random nodal values
    init_field = np.random.rand(Nx, Ny)
    padded_field = np.pad(init_field, (0, 1), mode="wrap")

    # Initialize field
    if use_device and HAS_CUPY:
        nodal_field.pg[...] = cp.asarray(padded_field)
    else:
        nodal_field.pg[...] = padded_field

    # Create operator
    kernel = get_quad_triangle_kernel()
    op_mugrid = muGrid.ConvolutionOperator([0, 0], kernel)

    device_str = nodal_field_cpp.device

    # Time the operation
    if use_device and HAS_CUPY:
        cp.cuda.Device().synchronize()
        start_event = cp.cuda.Event()
        end_event = cp.cuda.Event()
        start_event.record()

    t_start = time.perf_counter()
    op_mugrid.apply(nodal_field_cpp, quad_field_cpp)
    t_elapsed = time.perf_counter() - t_start

    if use_device and HAS_CUPY:
        end_event.record()
        end_event.synchronize()
        t_elapsed = cp.cuda.get_elapsed_time(start_event, end_event) / 1000.0

    # Get result back to host
    quad_field = wrap_field(quad_field_cpp)
    if use_device and HAS_CUPY:
        result = cp.asnumpy(quad_field.p)
    else:
        result = quad_field.p.copy()

    return result, t_elapsed, device_str, init_field


def quad_manual_combined(a, Nx, Ny, m_quad):
    """Compute quadrature manually using tensor-matrix multiplication."""
    F = a.reshape((Nx, Ny))

    # periodic wrap
    F = np.vstack([F, F[0:1, :]])
    F = np.hstack([F, F[:, 0:1]])

    # stack square-adjacent nodal values
    bl = F[:-1, :-1]
    br = F[:-1, 1:]
    tl = F[1:, :-1]
    tr = F[1:, 1:]

    stack = np.stack([bl, tl, br, tr], axis=-1)
    res = np.einsum("ij,xyj->xyi", m_quad, stack)
    return res  # shape (Nx, Ny, 6)


def test_quad_triangle_3_mugrid_vs_manual():
    """
    Compares wall-clock time and correctness of computing quadrature points on
    triangles using muGrid convolution vs. 'manual' tensor-matrix multiplication.
    """
    Nx, Ny = 1000, 1000
    nb_grid_pts = (Nx, Ny)

    # Run muGrid
    quad_mugrid, t_mugrid, device_str, init_field = run_quad_triangle_mugrid(
        nb_grid_pts, use_device=False
    )

    # Run manual
    m_quad = get_quad_triangle_manual_matrix()
    t_manual_start = time.perf_counter()
    quad_manual = quad_manual_combined(init_field, Nx, Ny, m_quad)
    t_manual = time.perf_counter() - t_manual_start

    quad_manual = quad_manual.transpose(2, 0, 1)  # -> shape (6, Nx, Ny)

    # Print timing
    print(f"muGrid operator time ({device_str}):  {t_mugrid:.6f} s")
    print(f"Manual operator time:  {t_manual:.6f} s")

    # Correctness check
    np.testing.assert_allclose(quad_mugrid, quad_manual, rtol=1e-10, atol=1e-12)

    # Compare wall-clock time
    # assert t_mugrid < 1.05 * t_manual, (
    #     f"muGrid slower than manual quadrature: "
    #     f"muGrid {t_mugrid:.6f}s vs. tensor-matrix mul {t_manual:.6f}s"
    # )


@unittest.skipUnless(GPU_AVAILABLE and HAS_CUPY, get_gpu_cupy_skip_reason() or "")
def test_quad_triangle_device_vs_host():
    """Compare quadrature triangle computation between CPU and GPU."""
    Nx, Ny = 1000, 1000
    nb_grid_pts = (Nx, Ny)

    # We need to use the same random seed for both runs
    np.random.seed(42)

    # Run on host
    host_result, t_host, host_device, init_field_host = run_quad_triangle_mugrid(
        nb_grid_pts, use_device=False
    )

    # Run on device with same input
    # Recreate the field collection and use same init_field
    if HAS_CUPY:
        memory_location = muGrid.GlobalFieldCollection.MemoryLocation.Device
    else:
        memory_location = muGrid.GlobalFieldCollection.MemoryLocation.Host

    fc = muGrid.GlobalFieldCollection(
        nb_grid_pts,
        sub_pts={"quad": 6},
        nb_ghosts_right=(1, 1),
        memory_location=memory_location,
    )
    nodal_field_cpp = fc.real_field("nodal")
    quad_field_cpp = fc.real_field("quad", 1, "quad")
    nodal_field = wrap_field(nodal_field_cpp)

    padded_field = np.pad(init_field_host, (0, 1), mode="wrap")
    nodal_field.pg[...] = cp.asarray(padded_field)

    kernel = get_quad_triangle_kernel()
    op_mugrid = muGrid.ConvolutionOperator([0, 0], kernel)
    device_str = nodal_field_cpp.device

    # Time device operation
    cp.cuda.Device().synchronize()
    start_event = cp.cuda.Event()
    end_event = cp.cuda.Event()
    start_event.record()
    op_mugrid.apply(nodal_field_cpp, quad_field_cpp)
    end_event.record()
    end_event.synchronize()
    t_device = cp.cuda.get_elapsed_time(start_event, end_event) / 1000.0

    quad_field = wrap_field(quad_field_cpp)
    device_result = cp.asnumpy(quad_field.p)

    # Check results match
    np.testing.assert_allclose(
        device_result,
        host_result,
        rtol=1e-10,
        atol=1e-12,
        err_msg="Device and host quadrature results differ",
    )

    # Print comparison
    speedup = t_host / t_device if t_device > 0 else 0
    print(f"\nQuadrature triangle comparison ({Nx}x{Ny}):")
    print(f"  Host ({host_device}):   {t_host:.6f} s")
    print(f"  Device ({device_str}): {t_device:.6f} s")
    print(f"  Speedup: {speedup:.2f}x")


@unittest.skipUnless(GPU_AVAILABLE and HAS_CUPY, get_gpu_cupy_skip_reason() or "")
def test_quad_triangle_device_scaling():
    """Test quadrature triangle scaling on GPU across different grid sizes."""
    grid_sizes = [256, 512, 1000, 2000]

    print("\n=== Quadrature Triangle GPU Scaling ===")
    print(f"{'Grid':>12} {'Host (s)':>12} {'Device (s)':>12} {'Speedup':>10}")
    print("-" * 50)

    for size in grid_sizes:
        nb_grid_pts = (size, size)

        # Host timing
        _, t_host, _, _ = run_quad_triangle_mugrid(nb_grid_pts, use_device=False)

        # Device timing
        _, t_device, device_str, _ = run_quad_triangle_mugrid(
            nb_grid_pts, use_device=True
        )

        speedup = t_host / t_device if t_device > 0 else 0
        print(f"{size}x{size:>5} {t_host:>12.6f} {t_device:>12.6f} {speedup:>10.2f}x")

    print("-" * 50)
    print(f"Device: {device_str}")


@unittest.skipUnless(GPU_AVAILABLE, get_gpu_skip_reason() or "")
class DeviceConvolutionPerformanceTests(unittest.TestCase):
    """Performance test suite for device (GPU) convolution operator"""

    def test_device_performance_small_grid(self):
        """Benchmark convolution on a small 32x32 grid on GPU"""
        print("\n=== Testing 2D Device Convolution Performance (Small Grid) ===")

        # Parameters
        nb_x_pts = 32
        nb_y_pts = 32
        nb_stencil_x = 3
        nb_stencil_y = 3
        nb_operators = 2
        nb_quad_pts = 4
        nb_field_components = 3

        # Create stencil
        stencil = np.random.rand(nb_operators, nb_quad_pts, nb_stencil_x, nb_stencil_y)

        conv_op = muGrid.ConvolutionOperator([-1, -1], stencil)

        # Create device field collection
        nb_ghosts = (1, 1)
        fc = muGrid.GlobalFieldCollection(
            (nb_x_pts, nb_y_pts),
            sub_pts={"quad": nb_quad_pts},
            nb_ghosts_left=nb_ghosts,
            nb_ghosts_right=nb_ghosts,
            memory_location=muGrid.GlobalFieldCollection.MemoryLocation.Device,
        )

        # Create fields
        nodal_cpp = fc.real_field("nodal", nb_field_components)
        quad_cpp = fc.real_field("quad", (nb_field_components, nb_operators), "quad")

        # Initialize with random data (transfer to device via CuPy if available)
        if HAS_CUPY:
            nodal_arr = cp.from_dlpack(nodal_cpp)
            nodal_shape = nodal_arr.shape
            # Create random data with correct shape
            random_data = cp.random.rand(*nodal_shape).astype(cp.float64)
            nodal_arr[...] = random_data

        device_str = nodal_cpp.device

        # Measure performance
        metrics = measure_convolution_performance(
            conv_op,
            nodal_cpp,
            quad_cpp,
            grid_size=(nb_x_pts, nb_y_pts),
            stencil_size=(nb_stencil_x, nb_stencil_y),
            num_iterations=100,
            device=device_str,
        )

        metrics.print_report()

        # Sanity checks
        self.assertGreater(metrics.total_flops, 0)
        self.assertGreater(metrics.gflops, 0.0)

    def test_device_performance_large_grid(self):
        """Benchmark convolution on a large 256x256 grid on GPU"""
        print("\n=== Testing 2D Device Convolution Performance (Large Grid) ===")

        # Parameters
        nb_x_pts = 256
        nb_y_pts = 256
        nb_stencil_x = 3
        nb_stencil_y = 3
        nb_operators = 2
        nb_quad_pts = 4
        nb_field_components = 3

        # Create stencil
        stencil = np.random.rand(
            nb_operators, nb_quad_pts, 1, nb_stencil_x, nb_stencil_y
        )

        conv_op = muGrid.ConvolutionOperator([-1, -1], stencil)

        # Create device field collection
        nb_ghosts = (1, 1)
        fc = muGrid.GlobalFieldCollection(
            (nb_x_pts, nb_y_pts),
            sub_pts={"quad": nb_quad_pts},
            nb_ghosts_left=nb_ghosts,
            nb_ghosts_right=nb_ghosts,
            memory_location=muGrid.GlobalFieldCollection.MemoryLocation.Device,
        )

        # Create fields
        nodal_cpp = fc.real_field("nodal", nb_field_components)
        quad_cpp = fc.real_field("quad", (nb_field_components, nb_operators), "quad")

        # Initialize with random data
        if HAS_CUPY:
            nodal_arr = cp.from_dlpack(nodal_cpp)
            nodal_arr[...] = cp.random.rand(*nodal_arr.shape).astype(cp.float64)

        device_str = nodal_cpp.device

        # Measure performance (fewer iterations for large grid)
        metrics = measure_convolution_performance(
            conv_op,
            nodal_cpp,
            quad_cpp,
            grid_size=(nb_x_pts, nb_y_pts),
            stencil_size=(nb_stencil_x, nb_stencil_y),
            num_iterations=100,
            device=device_str,
        )

        metrics.print_report()

        # Sanity checks
        self.assertGreater(metrics.total_flops, 0)
        self.assertGreater(metrics.gflops, 0.0)


@unittest.skipUnless(GPU_AVAILABLE and HAS_CUPY, get_gpu_cupy_skip_reason() or "")
class HostDeviceComparisonTests(unittest.TestCase):
    """Compare performance between host and device convolution"""

    def test_host_vs_device_performance(self):
        """Compare convolution performance between CPU and GPU"""
        print("\n=== Host vs Device Performance Comparison ===")

        grid_sizes = [32, 64, 128, 256]
        host_results = []
        device_results = []

        for grid_size in grid_sizes:
            nb_stencil = 3
            nb_operators = 2
            nb_quad_pts = 4
            nb_components = 3
            nb_ghosts = (1, 1)

            # Create stencil
            stencil = np.random.rand(nb_operators, nb_quad_pts, nb_stencil, nb_stencil)
            conv_op = muGrid.ConvolutionOperator([-1, -1], stencil)

            # Host field collection
            fc_host = muGrid.GlobalFieldCollection(
                (grid_size, grid_size),
                sub_pts={"quad": nb_quad_pts},
                nb_ghosts_left=nb_ghosts,
                nb_ghosts_right=nb_ghosts,
            )
            nodal_host = fc_host.real_field("nodal", nb_components)
            quad_host = fc_host.real_field(
                "quad", (nb_components, nb_operators), "quad"
            )

            # Initialize host field
            nodal_arr = np.from_dlpack(nodal_host)
            nodal_arr[...] = np.random.rand(*nodal_arr.shape)

            # Measure host performance
            iterations = 100 if grid_size <= 64 else 50
            host_metrics = measure_convolution_performance(
                conv_op,
                nodal_host,
                quad_host,
                grid_size=(grid_size, grid_size),
                stencil_size=(nb_stencil, nb_stencil),
                num_iterations=iterations,
                device="cpu",
            )
            host_results.append(host_metrics)

            # Device field collection
            fc_device = muGrid.GlobalFieldCollection(
                (grid_size, grid_size),
                sub_pts={"quad": nb_quad_pts},
                nb_ghosts_left=nb_ghosts,
                nb_ghosts_right=nb_ghosts,
                memory_location=muGrid.GlobalFieldCollection.MemoryLocation.Device,
            )
            nodal_device = fc_device.real_field("nodal", nb_components)
            quad_device = fc_device.real_field(
                "quad", (nb_components, nb_operators), "quad"
            )

            # Initialize device field
            nodal_device_arr = cp.from_dlpack(nodal_device)
            nodal_device_arr[...] = cp.random.rand(*nodal_device_arr.shape).astype(
                cp.float64
            )

            device_str = nodal_device.device

            # Measure device performance
            device_metrics = measure_convolution_performance(
                conv_op,
                nodal_device,
                quad_device,
                grid_size=(grid_size, grid_size),
                stencil_size=(nb_stencil, nb_stencil),
                num_iterations=iterations,
                device=device_str,
            )
            device_results.append(device_metrics)

        # Print comparison table
        print("\n" + "=" * 80)
        print(f"{'Grid':>10} {'Host GFLOPS':>15} {'Device GFLOPS':>15} {'Speedup':>12}")
        print("-" * 80)

        for host_m, device_m in zip(host_results, device_results):
            grid_pixels = np.prod(host_m.grid_size)
            speedup = device_m.gflops / host_m.gflops if host_m.gflops > 0 else 0

            print(
                f"{grid_pixels:>10} {host_m.gflops:>15.3f} "
                f"{device_m.gflops:>15.3f} {speedup:>12.2f}x"
            )

        print("=" * 80 + "\n")

    def test_device_convolution_correctness(self):
        """Verify that device convolution produces same results as host"""
        print("\n=== Device Convolution Correctness Check ===")

        # Parameters
        nb_x_pts = 16
        nb_y_pts = 16
        nb_stencil = 3
        nb_operators = 2
        nb_quad_pts = 4
        nb_components = 3
        nb_ghosts = (1, 1)

        # Create stencil
        stencil = np.random.rand(nb_operators, nb_quad_pts, nb_stencil, nb_stencil)
        conv_op = muGrid.ConvolutionOperator([-1, -1], stencil)

        # Create host fields
        fc_host = muGrid.GlobalFieldCollection(
            (nb_x_pts, nb_y_pts),
            sub_pts={"quad": nb_quad_pts},
            nb_ghosts_left=nb_ghosts,
            nb_ghosts_right=nb_ghosts,
        )
        nodal_host = fc_host.real_field("nodal", nb_components)
        quad_host = fc_host.real_field("quad", (nb_components, nb_operators), "quad")

        # Initialize host nodal field with known values
        nodal_host_arr = np.from_dlpack(nodal_host)
        test_data = np.random.rand(*nodal_host_arr.shape)
        nodal_host_arr[...] = test_data

        # Apply on host
        conv_op.apply(nodal_host, quad_host)
        host_result = np.from_dlpack(quad_host).copy()

        # Create device fields
        fc_device = muGrid.GlobalFieldCollection(
            (nb_x_pts, nb_y_pts),
            sub_pts={"quad": nb_quad_pts},
            nb_ghosts_left=nb_ghosts,
            nb_ghosts_right=nb_ghosts,
            memory_location=muGrid.GlobalFieldCollection.MemoryLocation.Device,
        )
        nodal_device = fc_device.real_field("nodal", nb_components)
        quad_device = fc_device.real_field(
            "quad", (nb_components, nb_operators), "quad"
        )

        # Copy same data to device
        nodal_device_arr = cp.from_dlpack(nodal_device)
        nodal_device_arr[...] = cp.asarray(test_data)

        # Apply on device
        conv_op.apply(nodal_device, quad_device)

        # Get device result back to host
        quad_device_arr = cp.from_dlpack(quad_device)
        device_result = cp.asnumpy(quad_device_arr)

        # Compare results
        np.testing.assert_allclose(
            device_result,
            host_result,
            rtol=1e-10,
            atol=1e-12,
            err_msg="Device convolution result differs from host result",
        )
        print("Device and host results match!")



import pytest

import muGrid

@pytest.fixture
def comm():
    try:
        from mpi4py import MPI
        return muGrid.Communicator(MPI.COMM_WORLD)
    except ImportError:
        return muGrid.Communicator()



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
"""
@file     python_field_tests.py

@author  Lars Pastewka <lars.pastewka@imtek.uni-freiburg.de>

@date    18 Oct 2019

@brief   test field collections and fields

Copyright © 2018 Till Junge

µGrid is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3, or (at
your option) any later version.

µGrid is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with µGrid; see the file COPYING. If not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.

Additional permission under GNU GPL version 3 section 7

If you modify this Program, or any covered work, by linking or combining it
with proprietary FFT implementations or numerical libraries, containing parts
covered by the terms of those libraries' licenses, the licensors of this
Program grant you additional permission to convey the resulting work.
"""

import unittest

import numpy as np

import muGrid
from muGrid import real_field, wrap_field


class FieldCheck(unittest.TestCase):
    def setUp(self):
        self.nb_grid_pts = (10, 11, 21)

    def test_buffer_size_one_quad_pt(self):
        values = np.random.rand(*self.nb_grid_pts)

        nb_grid_pts2 = (4,) + self.nb_grid_pts
        values2 = np.random.rand(*nb_grid_pts2)

        fc = muGrid.GlobalFieldCollection(self.nb_grid_pts)

        # Scalar - wrap field to access p/s/pg/sg properties
        fs_cpp = fc.register_real_field("test-field-scalar")
        fs = wrap_field(fs_cpp)
        self.assertEqual(fs.p.shape, self.nb_grid_pts)
        self.assertEqual(fs.s.shape, (1,) + self.nb_grid_pts)

        # Single component
        f_cpp = fc.register_real_field("test-field", 1)
        f = wrap_field(f_cpp)
        self.assertEqual(f.p.shape, (1,) + self.nb_grid_pts)
        self.assertEqual(f.s.shape, (1, 1) + self.nb_grid_pts)

        # Four components
        f2_cpp = fc.register_real_field("test-field2", 4)
        f2 = wrap_field(f2_cpp)
        self.assertEqual(f2.p.shape, nb_grid_pts2)
        self.assertEqual(f2.s.shape, (4, 1) + self.nb_grid_pts)

        # Check that we get those fields back
        self.assertEqual(fc.get_field('test-field-scalar'), fs_cpp)
        self.assertEqual(fc.get_field('test-field'), f_cpp)
        self.assertEqual(fc.get_field('test-field2'), f2_cpp)

        # Check strides
        self.assertEqual(fs.stride(muGrid.Pixel), 1)
        self.assertEqual(f.stride(muGrid.Pixel), 1)
        self.assertEqual(f2.stride(muGrid.Pixel), 4)
        self.assertEqual(fs.stride(muGrid.SubPt), 1)
        self.assertEqual(f.stride(muGrid.SubPt), 1)
        self.assertEqual(f2.stride(muGrid.SubPt), 4)

        # Check subpoint-shaped convenience access (read) and data assignment
        self.assertEqual(fs.s.shape, (1,) + self.nb_grid_pts)
        fs.s[...] = values.reshape((1,) + self.nb_grid_pts)
        np.testing.assert_allclose(fs.s, values.reshape((1,) + self.nb_grid_pts))

        self.assertEqual(f.s.shape, (1, 1) + self.nb_grid_pts)
        f.s[...] = values.reshape((1, 1) + self.nb_grid_pts)
        np.testing.assert_allclose(f.s, values.reshape((1, 1) + self.nb_grid_pts))

        self.assertEqual(f2.s.shape, (4, 1) + self.nb_grid_pts)

        # Check pixel-shaped convenience access (read) and data assignment
        self.assertEqual(fs.p.shape, self.nb_grid_pts)
        fs.p[...] = values
        np.testing.assert_allclose(fs.p, values)

        self.assertEqual(f.p.shape, (1,) + self.nb_grid_pts)
        f.p[...] = values.reshape((1,) + self.nb_grid_pts)
        np.testing.assert_allclose(f.p, values.reshape((1,) + self.nb_grid_pts))

        self.assertEqual(f2.p.shape, nb_grid_pts2)
        f2.p[...] = values2
        np.testing.assert_allclose(f2.p, values2)

    def test_buffer_size_four_quad_pt(self):
        fc = muGrid.GlobalFieldCollection(self.nb_grid_pts, sub_pts={'quad': 4})
        # Single component
        f_cpp = fc.register_real_field("test-field", 1, 'quad')
        f = wrap_field(f_cpp)
        self.assertEqual(f.p.shape, (4,) + self.nb_grid_pts)
        self.assertEqual(f.s.shape, (1, 4) + self.nb_grid_pts)
        # Four components
        f2_cpp = fc.register_real_field("test-field2", 3, 'quad')
        f2 = wrap_field(f2_cpp)
        self.assertEqual(f2.p.shape, (4 * 3,) + self.nb_grid_pts)
        self.assertEqual(f2.s.shape, (3, 4) + self.nb_grid_pts)
        # Check that we get those fields back
        self.assertEqual(fc.get_field('test-field'), f_cpp)
        self.assertEqual(fc.get_field('test-field2'), f2_cpp)
        # Check strides
        self.assertEqual(f.stride(muGrid.Pixel), 4)
        self.assertEqual(f2.stride(muGrid.Pixel), 3 * 4)
        self.assertEqual(f.stride(muGrid.SubPt), 1)
        self.assertEqual(f2.stride(muGrid.SubPt), 3)

    def test_buffer_access(self):
        fc = muGrid.GlobalFieldCollection(self.nb_grid_pts, sub_pts={'quad': 4})
        # Single component
        f_cpp = fc.register_real_field("test-field", 1, 'quad')
        f = wrap_field(f_cpp)

        # Access via DLPack - arrays from sg (full buffer)
        arr1 = f.sg  # SubPt layout with ghosts (full buffer)
        self.assertFalse(arr1.flags.owndata)
        self.assertTrue(arr1.flags.f_contiguous)

        arr2 = f.p  # Pixel layout
        self.assertFalse(arr2.flags.owndata)
        self.assertTrue(arr2.flags.f_contiguous)

        arr1[0, 0, 2, 1, 0] = 10  # f-contiguous
        self.assertEqual(arr2[0, 2, 1, 0], 10)  # f-contiguous

        f2_cpp = fc.register_real_field("test-field-2", 3, 'quad')
        f2 = wrap_field(f2_cpp)
        # Different fields should have different data buffers
        self.assertTrue(np.from_dlpack(f_cpp).ctypes.data !=
                        np.from_dlpack(f2_cpp).ctypes.data)

    def test_col_major(self):
        dims = (3, 4)
        fc = muGrid.GlobalFieldCollection(self.nb_grid_pts, self.nb_grid_pts,
                                          (0,) * len(self.nb_grid_pts),
                                          {})
        f_cpp = fc.register_real_field("test-field", dims)
        a = np.from_dlpack(f_cpp)
        self.assertTrue(a.flags.f_contiguous)
        self.assertEqual(a.shape, tuple(list(dims) + [1, ] + list(self.nb_grid_pts)))
        strides = np.append([1], np.cumprod(dims))
        strides = np.append(strides, strides[-1])
        strides = np.append(strides,
                            strides[-1] * np.cumprod(self.nb_grid_pts))
        strides = 8 * strides[:-1]
        self.assertEqual(a.strides, tuple(strides))

    def test_row_major(self):
        # We need to specify the storage order twice and they mean different
        # things.
        dims = (3, 4)
        fc = muGrid.GlobalFieldCollection(
            self.nb_grid_pts, self.nb_grid_pts, [0] * len(self.nb_grid_pts),
            muGrid.StorageOrder.RowMajor, {}, muGrid.StorageOrder.RowMajor)
        f_cpp = fc.register_real_field("test-field", dims)
        a = np.from_dlpack(f_cpp)
        self.assertTrue(a.flags.c_contiguous)
        self.assertEqual(a.shape, tuple(list(dims) + [1, ] + list(self.nb_grid_pts)))
        strides = np.append([1], np.cumprod(self.nb_grid_pts[::-1]))
        strides = np.append(strides, strides[-1])
        strides = np.append(strides, strides[-1] * np.cumprod(dims[::-1]))
        strides = 8 * strides[-2::-1]
        self.assertEqual(a.strides, tuple(strides))

    def test_local_field_collection(self):
        lfc_name = "ArbitraryLocalFieldCollectionName"
        lfc_field_name = "MyLocalField"
        fc = muGrid.LocalFieldCollection(3, lfc_name)
        fc.add_pixel(1)  # add pixel with global_index = 1
        fc.initialise()
        field_cpp = fc.register_real_field(lfc_field_name, 1, 'pixel')
        field_array = np.from_dlpack(field_cpp)
        field_array[:] = 5  # assigne value 5 to pixel 1

        self.assertEqual(fc.get_name(), lfc_name)

    def test_accessors(self):
        fc = muGrid.GlobalFieldCollection(self.nb_grid_pts)
        field_cpp = fc.register_real_field("test-field", 1, 'pixel')
        field = wrap_field(field_cpp)
        # Without ghosts, p and pg should have the same shape
        # Similarly for s and sg
        assert field.p.shape == field.pg.shape
        assert field.s.shape == field.sg.shape

    def test_dlpack_writability(self):
        """Test that arrays obtained via DLPack are writable."""
        fc = muGrid.GlobalFieldCollection(self.nb_grid_pts)
        field_cpp = fc.register_real_field("test-field", 1, 'pixel')
        arr = np.from_dlpack(field_cpp)

        # Should be writable
        self.assertTrue(arr.flags.writeable)

        # Test actual write
        arr[0, 0, 0, 0, 0] = 42.0
        self.assertEqual(arr[0, 0, 0, 0, 0], 42.0)

    def test_field_wrapper_properties(self):
        """Test the Python Field wrapper properties."""
        fc = muGrid.GlobalFieldCollection(self.nb_grid_pts)
        field_cpp = fc.register_real_field("test-field", 3, 'pixel')
        field = wrap_field(field_cpp)

        # Check that wrapper provides all expected properties
        self.assertEqual(field.name, "test-field")
        self.assertEqual(field.nb_components, 3)

        # Check array access
        self.assertEqual(field.sg.shape, (3, 1) + self.nb_grid_pts)
        self.assertEqual(field.s.shape, (3, 1) + self.nb_grid_pts)
        self.assertEqual(field.pg.shape, (3,) + self.nb_grid_pts)
        self.assertEqual(field.p.shape, (3,) + self.nb_grid_pts)


if __name__ == "__main__":
    unittest.main()



# Solver for the Poisson equation

import argparse
import time

try:
    import matplotlib.pyplot as plt
except ModuleNotFoundError:
    plt = None
import numpy as np

import muGrid
from muGrid import real_field
from muGrid.Solvers import conjugate_gradients

try:
    from mpi4py import MPI

    comm = muGrid.Communicator(MPI.COMM_WORLD)
except ImportError:
    comm = muGrid.Communicator()

from NuMPI.Testing.Subdivision import suggest_subdivisions

parser = argparse.ArgumentParser(
    prog="Poisson",
    description="Solve the Poisson equation"
)

parser.add_argument(
    "-n", "--nb-grid-pts",
    default=[32, 32],
    type=lambda s: [int(x) for x in s.split(",")],
    help="Grid points as nx,ny (default: 32,32)"
)

_memory_locations = {
    "host": muGrid.GlobalFieldCollection.MemoryLocation.Host,
    "device": muGrid.GlobalFieldCollection.MemoryLocation.Device,
}

parser.add_argument(
    "-m", "--memory",
    choices=_memory_locations,
    default="host",
    help="Memory space for allocation (default: host)"
)

parser.add_argument(
    "-i", "--maxiter",
    type=int,
    default=1000,
    help="Maximum number of CG iterations (default: 1000)"
)

parser.add_argument(
    "-p", "--plot",
    action="store_true",
    help="Show plot of RHS and solution (default: off)"
)

args = parser.parse_args()

if args.memory == "host":
    import numpy as arr
else:
    import cupy as arr

args.memory = _memory_locations[args.memory]

s = suggest_subdivisions(len(args.nb_grid_pts), comm.size)

decomposition = muGrid.CartesianDecomposition(comm, args.nb_grid_pts, s, (1, 1), (1, 1),
                                              memory_location=args.memory)
grid_spacing = 1 / np.array(args.nb_grid_pts)  # Grid spacing

stencil = np.array([[0, 1, 0], [1, -4, 1], [0, 1, 0]])  # FD-stencil for the Laplacian
laplace = muGrid.ConvolutionOperator([-1, -1], stencil)

x, y = decomposition.coords  # Domain-local coords for each pixel

# Create fields using the helper function - works directly with CartesianDecomposition
rhs = real_field(decomposition, "rhs")
solution = real_field(decomposition, "solution")

rhs.p[...] = arr.asarray((1 + np.cos(2 * np.pi * x) * np.cos(2 * np.pi * y)) ** 10)
rhs.p[...] -= arr.mean(rhs.p)


def callback(it, x, r, p):
    """
    Callback function to print the current solution, residual, and search direction.
    """
    print(f"{it:5} {arr.dot(r.ravel(), r.ravel()):.5}")


def hessp(x, Ax):
    """
    Function to compute the product of the Hessian matrix with a vector.
    The Hessian is represented by the convolution operator.
    """
    decomposition.communicate_ghosts(x._cpp)
    laplace.apply(x._cpp, Ax._cpp)
    # We need the minus sign because the Laplace operator is negative
    # definite, but the conjugate-gradients solver assumes a
    # positive-definite operator.
    Ax.s[...] /= -np.mean(grid_spacing) ** 2  # Scale by grid spacing
    return Ax


start_time = time.perf_counter()
try:
    conjugate_gradients(
        comm,
        decomposition.collection,
        hessp,  # linear operator
        rhs._cpp,  # Pass the underlying C++ field
        solution._cpp,
        tol=1e-6,
        callback=callback,
        maxiter=args.maxiter,
    )
    print("CG converged.")
except RuntimeError:
    print("CG did not converge.")
elapsed_time = time.perf_counter() - start_time
print(f"CG solver ran for {elapsed_time:.4f} seconds")

if args.plot:
    if plt is None:
        print("Warning: matplotlib not available, cannot show plot")
    else:
        fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(10, 5))
        ax1.imshow(rhs.p)
        ax1.set_title("RHS")
        ax2.imshow(solution.p)
        ax2.set_title("Solution")
        plt.tight_layout()
        plt.show()



import numpy as np
import muGrid

# Two dimensional grid
nb_grid_pts = (11, 12)
fc = muGrid.GlobalFieldCollection(nb_grid_pts)

# Get a tensor-field (for example to represent the strain)
strain = fc.real_field("strain", (2, 2))

# Fill the field with random numbers
strain.p[...] = np.random.rand(*((2, 2) + nb_grid_pts))



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
"""
@file   Field.py

@author Lars Pastewka <lars.pastewka@imtek.uni-freiburg.de>

@date   09 Dec 2024

@brief  Python wrapper for muGrid fields with numpy/cupy array views

Copyright © 2024 Lars Pastewka

µGrid is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3, or (at
your option) any later version.

µGrid is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with µGrid; see the file COPYING. If not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.

Additional permission under GNU GPL version 3 section 7

If you modify this Program, or any covered work, by linking or combining it
with proprietary FFT implementations or numerical libraries, containing parts
covered by the terms of those libraries' licenses, the licensors of this
Program grant you additional permission to convey the resulting work.
"""

import numpy as np

# CuPy is optional - only imported when GPU fields are used
_cupy = None


def _get_cupy():
    """Lazy import of CuPy for GPU array support."""
    global _cupy
    if _cupy is None:
        try:
            import cupy
            _cupy = cupy
        except ImportError:
            raise ImportError(
                "CuPy is required for GPU field access. "
                "Install it with: pip install cupy-cuda12x "
                "(or appropriate CUDA or ROCm version)"
            )
    return _cupy


class Field:
    """
    Python wrapper for muGrid fields providing numpy/cupy array views.

    This class wraps a C++ muGrid field and provides the following properties
    for accessing the underlying data as arrays:

    - `s`: SubPt layout, excluding ghost regions
    - `sg`: SubPt layout, including ghost regions
    - `p`: Pixel layout, excluding ghost regions
    - `pg`: Pixel layout, including ghost regions

    For CPU fields, the arrays are numpy arrays. For GPU fields (CUDA),
    the arrays are CuPy arrays. Both are views into the underlying C++ data,
    so modifications to the arrays will modify the field data directly (zero-copy).
    """

    def __init__(self, cpp_field):
        """
        Initialize the Field wrapper.

        Parameters
        ----------
        cpp_field : _muGrid.RealFieldBase or similar
            The underlying C++ field object
        """
        self._cpp = cpp_field
        self._buffer = None

    @property
    def is_on_gpu(self):
        """Check if this field resides on GPU memory."""
        return self._cpp.is_on_gpu

    @property
    def device(self):
        """Get the device where this field resides ('cpu' or 'cuda:N')."""
        return self._cpp.device

    def _get_buffer(self):
        """Lazy-load the full buffer via DLPack (requires numpy >= 2.1)."""
        if self._buffer is None:
            if self.is_on_gpu:
                cp = _get_cupy()
                self._buffer = cp.from_dlpack(self._cpp)
            else:
                self._buffer = np.from_dlpack(self._cpp)
        return self._buffer

    def _make_slice(self, offsets, shape):
        """Create slice tuple from offsets and shape."""
        return tuple(slice(o, o + s) for o, s in zip(offsets, shape))

    @property
    def sg(self):
        """
        SubPt layout array including ghost regions.

        Shape: (*components_shape, nb_sub_pts, *spatial_dims_with_ghosts)
        """
        return self._get_buffer()

    @property
    def s(self):
        """
        SubPt layout array excluding ghost regions.

        Shape: (*components_shape, nb_sub_pts, *spatial_dims_without_ghosts)
        """
        buf = self._get_buffer()
        offsets = self._cpp.offsets_s
        shape = self._cpp.shape_s
        slices = self._make_slice(offsets, shape)
        return buf[slices]

    @property
    def pg(self):
        """
        Pixel layout array including ghost regions.

        Shape: (nb_components * nb_sub_pts, *spatial_dims_with_ghosts)
        """
        buf = self._get_buffer()
        shape_pg = self._cpp.shape_pg
        return buf.reshape(shape_pg)

    @property
    def p(self):
        """
        Pixel layout array excluding ghost regions.

        Shape: (nb_components * nb_sub_pts, *spatial_dims_without_ghosts)
        """
        buf = self._get_buffer()
        shape_pg = self._cpp.shape_pg
        pixel_buf = buf.reshape(shape_pg)
        offsets = self._cpp.offsets_p
        shape = self._cpp.shape_p
        slices = self._make_slice(offsets, shape)
        return pixel_buf[slices]

    # Delegate attribute access to the underlying C++ object
    def __getattr__(self, name):
        """Delegate attribute access to the underlying C++ field."""
        return getattr(self._cpp, name)

    def __repr__(self):
        return f"Field({self._cpp.name!r}, shape={self._cpp.shape})"


def wrap_field(field):
    """
    Wrap a C++ field in a Python Field object.

    Parameters
    ----------
    field : _muGrid field object
        The underlying C++ field

    Returns
    -------
    Field
        Wrapped field with numpy array access
    """
    if isinstance(field, Field):
        return field
    return Field(field)


def real_field(collection, name, components=(), sub_pt="pixel"):
    """
    Create a real-valued field and return it wrapped with array accessors.

    The field is allocated in the collection's memory space (Host or Device).
    To create GPU fields, create the collection with memory_location=Device.

    Parameters
    ----------
    collection : FieldCollection or CartesianDecomposition
        The field collection or decomposition to create the field in.
        If a CartesianDecomposition is passed, uses its internal collection.
    name : str
        Name of the field
    components : tuple, optional
        Shape of the field components (default: scalar field)
    sub_pt : str, optional
        Sub-point type (default: "pixel")

    Returns
    -------
    Field
        Wrapped field with .s, .sg, .p, .pg accessors.
        Returns numpy arrays for host fields, CuPy arrays for device fields.
    """
    # Handle CartesianDecomposition by getting its collection
    fc = getattr(collection, 'collection', collection)
    cpp_field = fc.real_field(name, components, sub_pt)
    return Field(cpp_field)


def int_field(collection, name, components=(), sub_pt="pixel"):
    """
    Create an integer field and return it wrapped with array accessors.

    The field is allocated in the collection's memory space (Host or Device).
    To create GPU fields, create the collection with memory_location=Device.

    Parameters
    ----------
    collection : FieldCollection or CartesianDecomposition
        The field collection or decomposition to create the field in.
        If a CartesianDecomposition is passed, uses its internal collection.
    name : str
        Name of the field
    components : tuple, optional
        Shape of the field components (default: scalar field)
    sub_pt : str, optional
        Sub-point type (default: "pixel")

    Returns
    -------
    Field
        Wrapped field with .s, .sg, .p, .pg accessors.
        Returns numpy arrays for host fields, CuPy arrays for device fields.
    """
    # Handle CartesianDecomposition by getting its collection
    fc = getattr(collection, 'collection', collection)
    cpp_field = fc.int_field(name, components, sub_pt)
    return Field(cpp_field)


def uint_field(collection, name, components=(), sub_pt="pixel"):
    """
    Create an unsigned integer field and return it wrapped with array accessors.

    The field is allocated in the collection's memory space (Host or Device).
    To create GPU fields, create the collection with memory_location=Device.

    Parameters
    ----------
    collection : FieldCollection or CartesianDecomposition
        The field collection or decomposition to create the field in.
        If a CartesianDecomposition is passed, uses its internal collection.
    name : str
        Name of the field
    components : tuple, optional
        Shape of the field components (default: scalar field)
    sub_pt : str, optional
        Sub-point type (default: "pixel")

    Returns
    -------
    Field
        Wrapped field with .s, .sg, .p, .pg accessors.
        Returns numpy arrays for host fields, CuPy arrays for device fields.
    """
    # Handle CartesianDecomposition by getting its collection
    fc = getattr(collection, 'collection', collection)
    cpp_field = fc.uint_field(name, components, sub_pt)
    return Field(cpp_field)


def complex_field(collection, name, components=(), sub_pt="pixel"):
    """
    Create a complex-valued field and return it wrapped with array accessors.

    The field is allocated in the collection's memory space (Host or Device).
    To create GPU fields, create the collection with memory_location=Device.

    Parameters
    ----------
    collection : FieldCollection or CartesianDecomposition
        The field collection or decomposition to create the field in.
        If a CartesianDecomposition is passed, uses its internal collection.
    name : str
        Name of the field
    components : tuple, optional
        Shape of the field components (default: scalar field)
    sub_pt : str, optional
        Sub-point type (default: "pixel")

    Returns
    -------
    Field
        Wrapped field with .s, .sg, .p, .pg accessors.
        Returns numpy arrays for host fields, CuPy arrays for device fields.
    """
    # Handle CartesianDecomposition by getting its collection
    fc = getattr(collection, 'collection', collection)
    cpp_field = fc.complex_field(name, components, sub_pt)
    return Field(cpp_field)


# FFT-specific field creation functions

def fft_real_space_field(engine, name, nb_components=1):
    """
    Create a real-space field for FFT operations.

    This function creates a field in the FFT engine's real-space collection
    and returns it wrapped with array accessors.

    Parameters
    ----------
    engine : FFTEngine
        The FFT engine to create the field in
    name : str
        Name of the field
    nb_components : int, optional
        Number of field components (default: 1 for scalar field)

    Returns
    -------
    Field
        Wrapped real-valued field with .s, .sg, .p, .pg accessors
    """
    cpp_field = engine.real_space_field(name, nb_components)
    return Field(cpp_field)


def fft_fourier_space_field(engine, name, nb_components=1):
    """
    Create a Fourier-space field for FFT operations.

    This function creates a field in the FFT engine's Fourier-space collection
    and returns it wrapped with array accessors.

    Parameters
    ----------
    engine : FFTEngine
        The FFT engine to create the field in
    name : str
        Name of the field
    nb_components : int, optional
        Number of field components (default: 1 for scalar field)

    Returns
    -------
    Field
        Wrapped complex-valued field with .s, .sg, .p, .pg accessors
    """
    cpp_field = engine.fourier_space_field(name, nb_components)
    return Field(cpp_field)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
"""
@file   python_communicator_tests.py

@author Lars Pastewka <lars.pastewka@imtek.uni-freiburg.de>

@date   21 May 2019

@brief  Test muGrid's wrapper of the MPI communicator

Copyright © 2018 Till Junge

µGrid is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3, or (at
your option) any later version.

µGrid is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with µGrid; see the file COPYING. If not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.

Additional permission under GNU GPL version 3 section 7

If you modify this Program, or any covered work, by linking or combining it
with proprietary FFT implementations or numerical libraries, containing parts
covered by the terms of those libraries' licenses, the licensors of this
Program grant you additional permission to convey the resulting work.
"""

import numpy as np
import pytest

import muGrid


def test_sum_default():
    # The default communicator is COMM_SELF, i.e. each process by itself
    comm = muGrid.Communicator()
    assert comm.sum(comm.rank + 3) == 3


@pytest.mark.skipif(
    not muGrid.has_mpi, reason="muGrid was compiled without MPI support"
)
def test_sum_comm_world():
    try:
        from mpi4py import MPI

        comm = muGrid.Communicator(MPI.COMM_WORLD)
    except ImportError:
        comm = muGrid.Communicator()
    # 1 + 2 + 3 + ... + n = n*(n+1)/2
    assert comm.sum(comm.rank + 3) == comm.size * (comm.size + 1) / 2 + 2 * comm.size


def test_cum_sum_comm_world():
    try:
        from mpi4py import MPI

        comm = muGrid.Communicator(MPI.COMM_WORLD)
    except ImportError:
        comm = muGrid.Communicator()
    # 1 + 2 + 3 + ... + n = n*(n+1)/2
    assert (
        comm.cumulative_sum(comm.rank + 1)
        == comm.rank * (comm.rank + 1) / 2 + comm.rank + 1
    )


def test_bcast_1():
    # The default communicator is COMM_SELF, i.e. each process by itself
    comm = muGrid.Communicator()
    scalar_arg = comm.rank + 3
    res = comm.bcast(scalar_arg, 0)
    assert res == 3

    scalar_arg = comm.rank + 1
    res = comm.bcast(scalar_arg=scalar_arg, root=comm.size - 1)
    assert res == comm.size


# TODO(RLeute): I think we need such a test
# def test_get_comm():
#     # test if the function get comm works and returns the mpi communicator
#     from mpi4py import MPI
#     mpi_communicator = MPI.COMM_WORLD
#     comm = muGrid.Communicator(mpi_communicator)
#     self.assertEqual(comm.mpi_comm, mpi_communicator)


@pytest.mark.skipif(not muGrid.has_mpi, reason="muFFT was compiled without MPI support")
def test_bcast_2():
    try:
        from mpi4py import MPI

        comm = muGrid.Communicator(MPI.COMM_WORLD)
    except ImportError:
        comm = muGrid.Communicator()
    scalar_arg = comm.rank + 3
    res = comm.bcast(scalar_arg, 0)
    assert res == 3

    scalar_arg = comm.rank + 1
    res = comm.bcast(scalar_arg=scalar_arg, root=comm.size - 1)
    assert res == comm.size


def test_gather():
    try:
        from mpi4py import MPI

        comm = muGrid.Communicator(MPI.COMM_WORLD)
    except ImportError:
        comm = muGrid.Communicator()

    # gather arrays "a" with different lengths on the ranks
    a = np.arange(comm.rank * 2 + 4).reshape((-1, 2)).T

    a_gathered = comm.gather(a)

    # construct reference
    for i in range(comm.size):
        if i == 0:
            a_ref = np.arange(i * 2 + 4).reshape((-1, 2))
        elif i >= 1:
            a_tmp = np.arange(i * 2 + 4).reshape((-1, 2))
            a_ref = np.concatenate((a_ref, a_tmp), axis=0)

    assert (a_gathered == a_ref.T).all()



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
"""
@file     python_fft_tests.py

@author  Lars Pastewka <lars.pastewka@imtek.uni-freiburg.de>

@date    18 Dec 2024

@brief   test FFT engine Python bindings

Copyright © 2024 Lars Pastewka

µGrid is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3, or (at
your option) any later version.

µGrid is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with µGrid; see the file COPYING. If not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.

Additional permission under GNU GPL version 3 section 7

If you modify this Program, or any covered work, by linking or combining it
with proprietary FFT implementations or numerical libraries, containing parts
covered by the terms of those libraries' licenses, the licensors of this
Program grant you additional permission to convey the resulting work.
"""

import unittest

import numpy as np
from numpy.testing import assert_allclose, assert_array_equal

import muGrid
from muGrid import fft_fourier_space_field, fft_real_space_field


class FFTUtilsTest(unittest.TestCase):
    """Test FFT utility functions."""

    def test_fft_freqind_even(self):
        """Test fft_freqind for even size (should match numpy)."""
        freqs = muGrid.fft_freqind(8)
        expected = [0, 1, 2, 3, -4, -3, -2, -1]
        assert_array_equal(freqs, expected)

    def test_fft_freqind_odd(self):
        """Test fft_freqind for odd size (should match numpy)."""
        freqs = muGrid.fft_freqind(7)
        expected = [0, 1, 2, 3, -3, -2, -1]
        assert_array_equal(freqs, expected)

    def test_fft_freqind_matches_numpy(self):
        """Test that fft_freqind matches numpy.fft.fftfreq * n."""
        for n in [7, 8, 16, 17]:
            freqs = muGrid.fft_freqind(n)
            # numpy.fft.fftfreq returns floats, and astype(int) truncates.
            # Use round() for proper comparison.
            expected = np.round(np.fft.fftfreq(n) * n).astype(int)
            assert_array_equal(freqs, expected, err_msg=f"Failed for n={n}")

    def test_rfft_freqind(self):
        """Test rfft_freqind (half-complex)."""
        freqs = muGrid.rfft_freqind(8)
        expected = [0, 1, 2, 3, 4]
        assert_array_equal(freqs, expected)

    def test_rfft_freqind_matches_numpy(self):
        """Test that rfft_freqind matches numpy.fft.rfftfreq * n."""
        for n in [7, 8, 16, 17]:
            freqs = muGrid.rfft_freqind(n)
            # numpy.fft.rfftfreq returns floats, and astype(int) truncates.
            # Use round() for proper comparison.
            expected = np.round(np.fft.rfftfreq(n) * n).astype(int)
            assert_array_equal(freqs, expected, err_msg=f"Failed for n={n}")

    def test_fft_freq(self):
        """Test fft_freq values."""
        freqs = muGrid.fft_freq(8, 2.0)
        # freq = k / (N * dx) where k is frequency index
        expected = np.fft.fftfreq(8, 2.0)
        assert_allclose(freqs, expected, rtol=1e-10)

    def test_rfft_freq(self):
        """Test rfft_freq values."""
        freqs = muGrid.rfft_freq(8, 2.0)
        expected = np.fft.rfftfreq(8, 2.0)
        assert_allclose(freqs, expected, rtol=1e-10)

    def test_get_hermitian_grid_pts_2d(self):
        """Test hermitian grid dimensions for 2D."""
        fourier = muGrid.get_hermitian_grid_pts([8, 10])
        self.assertEqual(fourier[0], 5)  # 8/2 + 1
        self.assertEqual(fourier[1], 10)

    def test_get_hermitian_grid_pts_3d(self):
        """Test hermitian grid dimensions for 3D."""
        fourier = muGrid.get_hermitian_grid_pts([8, 10, 12])
        self.assertEqual(fourier[0], 5)  # 8/2 + 1
        self.assertEqual(fourier[1], 10)
        self.assertEqual(fourier[2], 12)

    def test_fft_normalization(self):
        """Test FFT normalization factor."""
        norm_2d = muGrid.fft_normalization([8, 10])
        self.assertAlmostEqual(norm_2d, 1.0 / 80, places=15)

        norm_3d = muGrid.fft_normalization([8, 10, 12])
        self.assertAlmostEqual(norm_3d, 1.0 / 960, places=15)


class FFTEngineCreationTest(unittest.TestCase):
    """Test FFT engine creation and properties."""

    def test_create_2d_engine(self):
        """Test creating a 2D FFT engine."""
        engine = muGrid.FFTEngine([16, 20])

        # Check real space dimensions
        self.assertEqual(engine.nb_domain_grid_pts[0], 16)
        self.assertEqual(engine.nb_domain_grid_pts[1], 20)

        # Check Fourier space dimensions (r2c transform)
        self.assertEqual(engine.nb_fourier_grid_pts[0], 9)  # 16/2 + 1
        self.assertEqual(engine.nb_fourier_grid_pts[1], 20)

    def test_create_3d_engine(self):
        """Test creating a 3D FFT engine."""
        engine = muGrid.FFTEngine([8, 10, 12])

        self.assertEqual(engine.nb_domain_grid_pts[0], 8)
        self.assertEqual(engine.nb_domain_grid_pts[1], 10)
        self.assertEqual(engine.nb_domain_grid_pts[2], 12)

        self.assertEqual(engine.nb_fourier_grid_pts[0], 5)  # 8/2 + 1
        self.assertEqual(engine.nb_fourier_grid_pts[1], 10)
        self.assertEqual(engine.nb_fourier_grid_pts[2], 12)

    def test_normalization(self):
        """Test normalization factor."""
        engine_2d = muGrid.FFTEngine([8, 10])
        self.assertAlmostEqual(engine_2d.normalisation, 1.0 / 80, places=15)

        engine_3d = muGrid.FFTEngine([8, 10, 12])
        self.assertAlmostEqual(engine_3d.normalisation, 1.0 / 960, places=15)

    def test_backend_name(self):
        """Test backend name is reported."""
        engine = muGrid.FFTEngine([8, 10])
        self.assertEqual(engine.backend_name, "PocketFFT")


class FFTFieldTest(unittest.TestCase):
    """Test FFT field creation and access."""

    def setUp(self):
        self.engine_2d = muGrid.FFTEngine([16, 20])
        self.engine_3d = muGrid.FFTEngine([8, 10, 12])

    def test_create_real_field(self):
        """Test creating real-space field."""
        field = fft_real_space_field(self.engine_2d, "test_real")
        self.assertEqual(field.name, "test_real")
        self.assertEqual(field.s.shape, (1, 1, 16, 20))

    def test_create_fourier_field(self):
        """Test creating Fourier-space field."""
        field = fft_fourier_space_field(self.engine_2d, "test_fourier")
        self.assertEqual(field.name, "test_fourier")
        self.assertEqual(field.s.shape, (1, 1, 9, 20))  # Half-complex

    def test_field_data_types(self):
        """Test field data types."""
        real_field = fft_real_space_field(self.engine_2d, "real")
        fourier_field = fft_fourier_space_field(self.engine_2d, "fourier")

        self.assertEqual(real_field.s.dtype, np.float64)
        self.assertEqual(fourier_field.s.dtype, np.complex128)

    def test_field_pixel_layout(self):
        """Test pixel layout access."""
        real_field = fft_real_space_field(self.engine_2d, "real_p")

        # SubPt layout: (components, sub_pts, nx, ny)
        self.assertEqual(real_field.s.shape, (1, 1, 16, 20))

        # Pixel layout: (components * sub_pts, nx, ny)
        self.assertEqual(real_field.p.shape, (1, 16, 20))

    def test_field_3d_shapes(self):
        """Test 3D field shapes."""
        real_field = fft_real_space_field(self.engine_3d, "real_3d")
        fourier_field = fft_fourier_space_field(self.engine_3d, "fourier_3d")

        self.assertEqual(real_field.s.shape, (1, 1, 8, 10, 12))
        self.assertEqual(fourier_field.s.shape, (1, 1, 5, 10, 12))

    def test_field_write_access(self):
        """Test that field data can be written."""
        field = fft_real_space_field(self.engine_2d, "writable")
        field.s[:] = 1.0
        self.assertTrue(np.all(field.s == 1.0))

        field.p[:] = 2.0
        self.assertTrue(np.all(field.p == 2.0))


class FFTRoundtripTest(unittest.TestCase):
    """Test FFT forward/inverse roundtrip."""

    def test_2d_roundtrip_sine(self):
        """Test 2D FFT roundtrip with sine wave."""
        engine = muGrid.FFTEngine([16, 20])
        real_field = fft_real_space_field(engine, "real")
        fourier_field = fft_fourier_space_field(engine, "fourier")

        # Initialize with sine wave
        x = np.arange(16 * 20).reshape(1, 1, 16, 20)
        real_field.s[:] = np.sin(2 * np.pi * x / (16 * 20))
        original = real_field.s.copy()

        # Forward and inverse FFT
        engine.fft(real_field._cpp, fourier_field._cpp)
        engine.ifft(fourier_field._cpp, real_field._cpp)

        # Normalize
        real_field.s[:] *= engine.normalisation

        # Check roundtrip
        assert_allclose(real_field.s, original, atol=1e-14)

    def test_2d_roundtrip_random(self):
        """Test 2D FFT roundtrip with random data."""
        engine = muGrid.FFTEngine([16, 20])
        real_field = fft_real_space_field(engine, "real")
        fourier_field = fft_fourier_space_field(engine, "fourier")

        # Initialize with random data
        np.random.seed(42)
        real_field.s[:] = np.random.randn(1, 1, 16, 20)
        original = real_field.s.copy()

        # Forward and inverse FFT
        engine.fft(real_field._cpp, fourier_field._cpp)
        engine.ifft(fourier_field._cpp, real_field._cpp)
        real_field.s[:] *= engine.normalisation

        # Check roundtrip
        assert_allclose(real_field.s, original, atol=1e-14)

    def test_3d_roundtrip(self):
        """Test 3D FFT roundtrip."""
        engine = muGrid.FFTEngine([8, 10, 12])
        real_field = fft_real_space_field(engine, "real")
        fourier_field = fft_fourier_space_field(engine, "fourier")

        # Initialize
        np.random.seed(42)
        real_field.s[:] = np.random.randn(1, 1, 8, 10, 12)
        original = real_field.s.copy()

        # Forward and inverse FFT
        engine.fft(real_field._cpp, fourier_field._cpp)
        engine.ifft(fourier_field._cpp, real_field._cpp)
        real_field.s[:] *= engine.normalisation

        # Check roundtrip
        assert_allclose(real_field.s, original, atol=1e-14)

    def test_fft_dc_component(self):
        """Test that constant field has only DC component."""
        engine = muGrid.FFTEngine([8, 10])
        real_field = fft_real_space_field(engine, "real")
        fourier_field = fft_fourier_space_field(engine, "fourier")

        # Constant field
        real_field.s[:] = 5.0

        engine.fft(real_field._cpp, fourier_field._cpp)

        # DC component should be sum = 5 * 80 = 400
        self.assertAlmostEqual(fourier_field.s[0, 0, 0, 0].real, 400.0, places=10)
        self.assertAlmostEqual(fourier_field.s[0, 0, 0, 0].imag, 0.0, places=10)

        # All other components should be zero
        fourier_copy = fourier_field.s.copy()
        fourier_copy[0, 0, 0, 0] = 0
        assert_allclose(fourier_copy, 0, atol=1e-12)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
"""
@file     python_mpi_fft_tests.py

@author  Lars Pastewka <lars.pastewka@imtek.uni-freiburg.de>

@date    19 Dec 2024

@brief   Test MPI parallel FFT engine Python bindings

Copyright © 2024 Lars Pastewka

µGrid is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3, or (at
your option) any later version.

µGrid is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with µGrid; see the file COPYING. If not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.

Additional permission under GNU GPL version 3 section 7

If you modify this Program, or any covered work, by linking or combining it
with proprietary FFT implementations or numerical libraries, containing parts
covered by the terms of those libraries' licenses, the licensors of this
Program grant you additional permission to convey the resulting work.
"""

import numpy as np
import pytest
from numpy.testing import assert_allclose

from muGrid import FFTEngine, fft_fourier_space_field, fft_real_space_field


def make_subdomain_slices(locations, nb_pts):
    """Create slice objects for extracting subdomain from global array.

    Parameters
    ----------
    locations : tuple
        Starting indices of subdomain
    nb_pts : tuple
        Number of points in subdomain

    Returns
    -------
    tuple of slices
        Slices for each dimension
    """
    return tuple(slice(loc, loc + n) for loc, n in zip(locations, nb_pts))


def make_fourier_slices(locations, nb_pts):
    """Create slice objects for extracting subdomain from global Fourier array.

    Parameters
    ----------
    locations : tuple
        Starting indices of Fourier subdomain
    nb_pts : tuple
        Number of points in Fourier subdomain

    Returns
    -------
    tuple of slices
        Slices for each dimension
    """
    return tuple(slice(loc, loc + n) for loc, n in zip(locations, nb_pts))


class TestMPIFFTEngineConstruction:
    """Test FFT engine construction with MPI."""

    def test_create_engine_with_communicator(self, comm):
        """Test that FFT engine can be created with an MPI communicator."""
        nb_grid_pts = [16 * max(1, comm.size), 20]
        engine = FFTEngine(nb_grid_pts, comm)

        # Check global dimensions
        assert engine.nb_domain_grid_pts[0] == nb_grid_pts[0]
        assert engine.nb_domain_grid_pts[1] == nb_grid_pts[1]

        # Check Fourier dimensions (r2c transform)
        assert engine.nb_fourier_grid_pts[0] == nb_grid_pts[0] // 2 + 1
        assert engine.nb_fourier_grid_pts[1] == nb_grid_pts[1]

    def test_subdomain_decomposition_sums_to_total(self, comm):
        """Test that subdomain sizes sum to total domain size."""
        nb_grid_pts = [16 * max(1, comm.size), 20]
        engine = FFTEngine(nb_grid_pts, comm)

        # Sum subdomain sizes across all ranks
        local_pts = np.prod(engine.nb_subdomain_grid_pts)
        total_pts = comm.sum(local_pts)
        assert total_pts == np.prod(nb_grid_pts)

    def test_subdomain_decomposition_3d(self, comm):
        """Test 3D subdomain decomposition."""
        nb_grid_pts = [8 * max(1, comm.size), 10, 12]
        engine = FFTEngine(nb_grid_pts, comm)

        # Sum subdomain sizes across all ranks
        local_pts = np.prod(engine.nb_subdomain_grid_pts)
        total_pts = comm.sum(local_pts)
        assert total_pts == np.prod(nb_grid_pts)

    def test_fourier_subdomain_decomposition_sums_to_total(self, comm):
        """Test that Fourier subdomain sizes sum to total Fourier domain."""
        nb_grid_pts = [16 * max(1, comm.size), 20]
        engine = FFTEngine(nb_grid_pts, comm)

        # Sum Fourier subdomain sizes across all ranks
        local_pts = np.prod(engine.nb_fourier_subdomain_grid_pts)
        total_pts = comm.sum(local_pts)
        assert total_pts == np.prod(engine.nb_fourier_grid_pts)

    def test_normalization_factor(self, comm):
        """Test that normalization factor is correct."""
        nb_grid_pts = [16, 20]
        engine = FFTEngine(nb_grid_pts, comm)
        expected_norm = 1.0 / np.prod(nb_grid_pts)
        assert abs(engine.normalisation - expected_norm) < 1e-15

    def test_backend_name(self, comm):
        """Test that backend name is reported."""
        nb_grid_pts = [16, 20]
        engine = FFTEngine(nb_grid_pts, comm)
        # Should be PocketFFT for serial or MPI version
        assert engine.backend_name == "PocketFFT"


class TestMPIFFTForwardTransform:
    """Test forward FFT with MPI parallelization."""

    @pytest.mark.parametrize(
        "nb_grid_pts",
        [
            [16, 20],
            [32, 24],
        ],
    )
    def test_forward_transform_2d(self, comm, nb_grid_pts):
        """Test 2D forward FFT against numpy reference."""
        # Scale grid with number of processes
        nb_grid_pts = [n * max(1, comm.size) for n in nb_grid_pts]
        engine = FFTEngine(nb_grid_pts, comm)

        real_field = fft_real_space_field(engine, "real")
        fourier_field = fft_fourier_space_field(engine, "fourier")

        # Create global input array
        np.random.seed(42)
        global_input = np.random.randn(*nb_grid_pts)

        # Compute global reference using numpy
        # numpy uses row-major, we use column-major, so transpose
        global_ref = np.fft.rfftn(global_input.T).T

        # Extract local subdomain
        subdomain_slices = make_subdomain_slices(
            engine.subdomain_locations, engine.nb_subdomain_grid_pts
        )
        local_input = global_input[subdomain_slices]

        # Set local field data (squeeze to remove component dimension)
        real_field.p[0, ...] = local_input

        # Perform FFT
        engine.fft(real_field._cpp, fourier_field._cpp)

        # Extract expected local Fourier data
        fourier_slices = make_fourier_slices(
            engine.fourier_subdomain_locations, engine.nb_fourier_subdomain_grid_pts
        )
        expected_local = global_ref[fourier_slices]

        # Compare (squeeze to remove component dimension)
        tol = 1e-12 * np.prod(nb_grid_pts)
        assert_allclose(fourier_field.p.squeeze(), expected_local, atol=tol)

    @pytest.mark.parametrize(
        "nb_grid_pts",
        [
            [8, 10, 12],
        ],
    )
    def test_forward_transform_3d(self, comm, nb_grid_pts):
        """Test 3D forward FFT against numpy reference."""
        nb_grid_pts = [n * max(1, comm.size) for n in nb_grid_pts]
        engine = FFTEngine(nb_grid_pts, comm)

        real_field = fft_real_space_field(engine, "real")
        fourier_field = fft_fourier_space_field(engine, "fourier")

        # Create global input array
        np.random.seed(42)
        global_input = np.random.randn(*nb_grid_pts)

        # Compute global reference using numpy
        global_ref = np.fft.rfftn(global_input.T).T

        # Extract local subdomain
        subdomain_slices = make_subdomain_slices(
            engine.subdomain_locations, engine.nb_subdomain_grid_pts
        )
        local_input = global_input[subdomain_slices]
        real_field.p[0, ...] = local_input

        # Perform FFT
        engine.fft(real_field._cpp, fourier_field._cpp)

        # Extract expected local Fourier data
        fourier_slices = make_fourier_slices(
            engine.fourier_subdomain_locations, engine.nb_fourier_subdomain_grid_pts
        )
        expected_local = global_ref[fourier_slices]

        tol = 1e-12 * np.prod(nb_grid_pts)
        assert_allclose(fourier_field.p.squeeze(), expected_local, atol=tol)


class TestMPIFFTInverseTransform:
    """Test inverse FFT with MPI parallelization."""

    @pytest.mark.parametrize(
        "nb_grid_pts",
        [
            [16, 20],
            [32, 24],
        ],
    )
    def test_inverse_transform_2d(self, comm, nb_grid_pts):
        """Test 2D inverse FFT against numpy reference."""
        nb_grid_pts = [n * max(1, comm.size) for n in nb_grid_pts]
        engine = FFTEngine(nb_grid_pts, comm)

        real_field = fft_real_space_field(engine, "real")
        fourier_field = fft_fourier_space_field(engine, "fourier")

        # Create global Fourier input array
        fourier_shape = list(engine.nb_fourier_grid_pts)
        np.random.seed(42)
        global_fourier = np.zeros(fourier_shape, dtype=complex)
        global_fourier.real = np.random.randn(*fourier_shape)
        global_fourier.imag = np.random.randn(*fourier_shape)

        # Compute global reference using numpy
        global_ref = np.fft.irfftn(global_fourier.T).T

        # Extract local Fourier subdomain
        fourier_slices = make_fourier_slices(
            engine.fourier_subdomain_locations, engine.nb_fourier_subdomain_grid_pts
        )
        local_fourier = global_fourier[fourier_slices]
        fourier_field.p[0, ...] = local_fourier

        # Perform inverse FFT
        engine.ifft(fourier_field._cpp, real_field._cpp)
        real_field.p[0, ...] *= engine.normalisation

        # Extract expected local real data
        subdomain_slices = make_subdomain_slices(
            engine.subdomain_locations, engine.nb_subdomain_grid_pts
        )
        expected_local = global_ref[subdomain_slices]

        tol = 1e-12 * np.prod(nb_grid_pts)
        assert_allclose(real_field.p.squeeze(), expected_local, atol=tol)


class TestMPIFFTRoundtrip:
    """Test FFT roundtrip (forward + inverse)."""

    @pytest.mark.parametrize(
        "nb_grid_pts",
        [
            [16, 20],
            [8, 10, 12],
        ],
    )
    def test_roundtrip(self, comm, nb_grid_pts):
        """Test that ifft(fft(x)) * normalisation == x."""
        nb_grid_pts = [n * max(1, comm.size) for n in nb_grid_pts]
        engine = FFTEngine(nb_grid_pts, comm)

        real_field = fft_real_space_field(engine, "real")
        fourier_field = fft_fourier_space_field(engine, "fourier")

        # Initialize with random data
        np.random.seed(42 + comm.rank)  # Different seed per rank
        original = np.random.randn(*engine.nb_subdomain_grid_pts)
        real_field.p[0, ...] = original

        # Forward FFT
        engine.fft(real_field._cpp, fourier_field._cpp)

        # Inverse FFT
        engine.ifft(fourier_field._cpp, real_field._cpp)

        # Normalize
        real_field.p[0, ...] *= engine.normalisation

        # Check roundtrip
        assert_allclose(real_field.p.squeeze(), original, atol=1e-14)

    def test_roundtrip_sine_wave(self, comm):
        """Test roundtrip with sine wave pattern."""
        nb_grid_pts = [32 * max(1, comm.size), 32]
        engine = FFTEngine(nb_grid_pts, comm)

        real_field = fft_real_space_field(engine, "real")
        fourier_field = fft_fourier_space_field(engine, "fourier")

        # Create sine wave
        x = np.arange(
            engine.subdomain_locations[0],
            engine.subdomain_locations[0] + engine.nb_subdomain_grid_pts[0],
        )
        y = np.arange(
            engine.subdomain_locations[1],
            engine.subdomain_locations[1] + engine.nb_subdomain_grid_pts[1],
        )
        X, Y = np.meshgrid(x, y, indexing="ij")

        original = np.sin(2 * np.pi * X / nb_grid_pts[0]) * np.cos(
            2 * np.pi * Y / nb_grid_pts[1]
        )
        real_field.p[0, ...] = original

        # Roundtrip
        engine.fft(real_field._cpp, fourier_field._cpp)
        engine.ifft(fourier_field._cpp, real_field._cpp)
        real_field.p[0, ...] *= engine.normalisation

        assert_allclose(real_field.p.squeeze(), original, atol=1e-13)


class TestMPIFFTMultipleComponents:
    """Test FFT with multi-component fields."""

    def test_vector_field_2d(self, comm):
        """Test FFT of 2D vector field (2 components)."""
        nb_grid_pts = [16 * max(1, comm.size), 20]
        engine = FFTEngine(nb_grid_pts, comm)

        # Create 2-component fields
        real_field = fft_real_space_field(engine, "real_vec", nb_components=2)
        fourier_field = fft_fourier_space_field(engine, "fourier_vec", nb_components=2)

        # Initialize with random data
        np.random.seed(42 + comm.rank)
        original = np.random.randn(2, *engine.nb_subdomain_grid_pts)
        real_field.p[:] = original

        # Roundtrip
        engine.fft(real_field._cpp, fourier_field._cpp)
        engine.ifft(fourier_field._cpp, real_field._cpp)
        real_field.p[:] *= engine.normalisation

        assert_allclose(real_field.p, original, atol=1e-14)

    def test_tensor_field_3d(self, comm):
        """Test FFT of 3x3 tensor field in 3D."""
        nb_grid_pts = [8, 8, 8]
        engine = FFTEngine(nb_grid_pts, comm)

        # Create 9-component field (3x3 tensor)
        real_field = fft_real_space_field(engine, "real_tensor", nb_components=9)
        fourier_field = fft_fourier_space_field(
            engine, "fourier_tensor", nb_components=9
        )

        # Initialize with random data
        np.random.seed(42)
        original = np.random.randn(9, *engine.nb_subdomain_grid_pts)
        real_field.p[:] = original

        # Roundtrip
        engine.fft(real_field._cpp, fourier_field._cpp)
        engine.ifft(fourier_field._cpp, real_field._cpp)
        real_field.p[:] *= engine.normalisation

        assert_allclose(real_field.p, original, atol=1e-14)


class TestMPIFFTDCComponent:
    """Test DC (zero frequency) component handling."""

    def test_constant_field_dc_component(self, comm):
        """Test that constant field has only DC component."""
        nb_grid_pts = [16 * max(1, comm.size), 20]
        engine = FFTEngine(nb_grid_pts, comm)

        real_field = fft_real_space_field(engine, "real")
        fourier_field = fft_fourier_space_field(engine, "fourier")

        # Constant field
        constant_value = 5.0
        real_field.p[0, ...] = constant_value

        engine.fft(real_field._cpp, fourier_field._cpp)

        # DC component should be at (0,0) = constant_value * total_points
        expected_dc = constant_value * np.prod(nb_grid_pts)

        # Check if this rank has the DC component
        if (
            engine.fourier_subdomain_locations[0] == 0
            and engine.fourier_subdomain_locations[1] == 0
        ):
            assert abs(fourier_field.p[0, 0, 0].real - expected_dc) < 1e-10
            assert abs(fourier_field.p[0, 0, 0].imag) < 1e-10

        # All other components should be zero (globally)
        local_non_dc_sum = 0.0
        for i in range(engine.nb_fourier_subdomain_grid_pts[0]):
            for j in range(engine.nb_fourier_subdomain_grid_pts[1]):
                gi = engine.fourier_subdomain_locations[0] + i
                gj = engine.fourier_subdomain_locations[1] + j
                if gi != 0 or gj != 0:
                    local_non_dc_sum += abs(fourier_field.p[0, i, j])

        global_non_dc_sum = comm.sum(local_non_dc_sum)
        assert global_non_dc_sum < 1e-10


class TestMPIFFTProcessGrid:
    """Test process grid properties."""

    def test_process_grid_matches_communicator_size(self, comm):
        """Test that process grid size matches communicator size."""
        nb_grid_pts = [16 * max(1, comm.size), 20]
        engine = FFTEngine(nb_grid_pts, comm)

        p1, p2 = engine.process_grid
        assert p1 * p2 == comm.size

    def test_process_coords_unique(self, comm):
        """Test that each rank has unique process coordinates."""
        if comm.size == 1:
            pytest.skip("Need multiple processes for this test")

        nb_grid_pts = [16 * comm.size, 20]
        engine = FFTEngine(nb_grid_pts, comm)

        # Gather all process coordinates
        local_coords = np.array(engine.process_coords, dtype=np.int32).reshape(1, -1)
        all_coords = comm.gather(local_coords)

        if comm.rank == 0:
            # Check all coordinates are unique
            # all_coords is concatenated to [comm.size * 2] array after gather
            # Reshape to [comm.size, 2]
            all_coords = all_coords.reshape(comm.size, 2)
            coord_tuples = [tuple(row) for row in all_coords]
            coord_set = set(coord_tuples)
            assert len(coord_set) == comm.size


class TestMPIFFTEdgeCases:
    """Test edge cases for MPI FFT."""

    def test_small_grid_large_process_count(self, comm):
        """Test behavior with grid smaller than optimal for process count."""
        if comm.size > 4:
            pytest.skip("Test designed for <= 4 processes")

        # Small grid that may not divide evenly
        nb_grid_pts = [7, 5]
        engine = FFTEngine(nb_grid_pts, comm)

        # Should still work correctly
        real_field = fft_real_space_field(engine, "real")
        fourier_field = fft_fourier_space_field(engine, "fourier")

        np.random.seed(42 + comm.rank)
        subdomain_shape = engine.nb_subdomain_grid_pts
        if np.prod(subdomain_shape) > 0:
            original = np.random.randn(*subdomain_shape)
            real_field.p[0, ...] = original

            engine.fft(real_field._cpp, fourier_field._cpp)
            engine.ifft(fourier_field._cpp, real_field._cpp)
            real_field.p[0, ...] *= engine.normalisation

            # Compare without squeeze to handle subdomains with size-1 dimensions
            assert_allclose(real_field.p[0, ...], original, atol=1e-14)

    def test_non_power_of_two_grid(self, comm):
        """Test FFT with non-power-of-two grid sizes."""
        nb_grid_pts = [17 * max(1, comm.size), 23]
        engine = FFTEngine(nb_grid_pts, comm)

        real_field = fft_real_space_field(engine, "real")
        fourier_field = fft_fourier_space_field(engine, "fourier")

        np.random.seed(42 + comm.rank)
        original = np.random.randn(*engine.nb_subdomain_grid_pts)
        real_field.p[0, ...] = original

        engine.fft(real_field._cpp, fourier_field._cpp)
        engine.ifft(fourier_field._cpp, real_field._cpp)
        real_field.p[0, ...] *= engine.normalisation

        assert_allclose(real_field.p.squeeze(), original, atol=1e-14)


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
"""
@file   __init__.py

@author Lars Pastewka <lars.pastewka@imtek.uni-freiburg.de>

@date   10 Oct 2019

@brief  Main entry point for muGrid Python module

Copyright © 2018 Till Junge

µGrid is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3, or (at
your option) any later version.

µGrid is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with µGrid; see the file COPYING. If not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.

Additional permission under GNU GPL version 3 section 7

If you modify this Program, or any covered work, by linking or combining it
with proprietary FFT implementations or numerical libraries, containing parts
covered by the terms of those libraries' licenses, the licensors of this
Program grant you additional permission to convey the resulting work.
"""

try:
    from mpi4py import MPI
except ModuleNotFoundError:
    MPI = None

# Import the C++ extension module
# Try relative import first (for installed wheels where _muGrid.so is in the package)
# Fall back to absolute import (for development where _muGrid.so is in build directory)
try:
    from . import _muGrid
except ImportError:
    import _muGrid

has_mpi = _muGrid.Communicator.has_mpi
if has_mpi and MPI is None:
    raise RuntimeError("MPI support is enabled for muGrid but mpi4py is not available.")

# Feature flags for compile-time configuration
has_cuda = _muGrid.has_cuda
has_rocm = _muGrid.has_rocm
has_gpu = _muGrid.has_gpu
has_netcdf = _muGrid.has_netcdf
host_arch = _muGrid.host_arch
device_arch = _muGrid.device_arch

# Import Python wrappers for main classes (these accept wrapped Field objects)
from .Wrappers import (  # noqa: E402
    CartesianDecomposition,
    ConvolutionOperator,
    FFTEngine,
    FileIONetCDF,
    GlobalFieldCollection,
    LocalFieldCollection,
)

# Expose OpenMode for FileIONetCDF if available
if hasattr(_muGrid, "FileIONetCDF"):
    OpenMode = _muGrid.FileIONetCDF.OpenMode

# Low-level C++ classes (for advanced use cases)
# These are prefixed with underscore to indicate they're internal
_CartesianDecomposition = _muGrid.CartesianDecomposition
_ConvolutionOperator = _muGrid.ConvolutionOperator
_FFTEngine = _muGrid.FFTEngine
_GlobalFieldCollection = _muGrid.GlobalFieldCollection
_LocalFieldCollection = _muGrid.LocalFieldCollection
if hasattr(_muGrid, "FileIONetCDF"):
    _FileIONetCDF = _muGrid.FileIONetCDF

# Base classes and utilities (always C++ objects)
ConvolutionOperatorBase = _muGrid.ConvolutionOperatorBase
Decomposition = _muGrid.Decomposition
DynCcoord = _muGrid.DynCcoord
DynRcoord = _muGrid.DynRcoord
IterUnit = _muGrid.IterUnit
Pixel = _muGrid.Pixel
StorageOrder = _muGrid.StorageOrder
SubPt = _muGrid.SubPt
Unit = _muGrid.Unit
Verbosity = _muGrid.Verbosity

# FFT utility functions
fft_freq = _muGrid.fft_freq
fft_freqind = _muGrid.fft_freqind
fft_normalization = _muGrid.fft_normalization
get_hermitian_grid_pts = _muGrid.get_hermitian_grid_pts
rfft_freq = _muGrid.rfft_freq
rfft_freqind = _muGrid.rfft_freqind

# Domain indexing utilities
get_domain_ccoord = _muGrid.get_domain_ccoord
get_domain_index = _muGrid.get_domain_index

# Field classes and utilities
from .Field import Field  # noqa: E402
from .Field import complex_field  # noqa: E402
from .Field import fft_fourier_space_field  # noqa: E402
from .Field import fft_real_space_field  # noqa: E402
from .Field import int_field  # noqa: E402
from .Field import real_field  # noqa: E402
from .Field import uint_field  # noqa: E402
from .Field import wrap_field  # noqa: E402

# MPI communicator
from .Parallel import Communicator  # noqa: E402

# Version information
__version__ = _muGrid.version.description()

# Define public API
__all__ = [
    # Feature flags
    "has_mpi",
    "has_cuda",
    "has_rocm",
    "has_gpu",
    "has_netcdf",
    "host_arch",
    "device_arch",
    # Main classes (Python wrappers)
    "CartesianDecomposition",
    "Communicator",
    "ConvolutionOperator",
    "FFTEngine",
    "Field",
    "FileIONetCDF",
    "GlobalFieldCollection",
    "LocalFieldCollection",
    # Field creation functions
    "complex_field",
    "fft_fourier_space_field",
    "fft_real_space_field",
    "int_field",
    "real_field",
    "uint_field",
    "wrap_field",
    # FFT utilities
    "fft_freq",
    "fft_freqind",
    "fft_normalization",
    "get_hermitian_grid_pts",
    "rfft_freq",
    "rfft_freqind",
    # Domain utilities
    "get_domain_ccoord",
    "get_domain_index",
    # Enums and types
    "ConvolutionOperatorBase",
    "Decomposition",
    "DynCcoord",
    "DynRcoord",
    "IterUnit",
    "OpenMode",
    "Pixel",
    "StorageOrder",
    "SubPt",
    "Unit",
    "Verbosity",
]



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
"""
@file    python_convolution_operator_tests.py

@author  Yizhen Wang <yizhen.wang@imtek.uni-freiburg.de>

@date    11 Dec 2024

@brief   test field collections and fields

Copyright © 2018 Till Junge

µGrid is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3, or (at
your option) any later version.

µGrid is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with µGrid; see the file COPYING. If not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.

Additional permission under GNU GPL version 3 section 7

If you modify this Program, or any covered work, by linking or combining it
with proprietary FFT implementations or numerical libraries, containing parts
covered by the terms of those libraries' licenses, the licensors of this
Program grant you additional permission to convey the resulting work.
"""

import functools
import unittest

import numpy as np
import pytest
from NuMPI.Testing.Subdivision import suggest_subdivisions

import muGrid
from muGrid import real_field


class ConvolutionOperatorCheck(unittest.TestCase):
    """Test suite for ConvolutionOperator functionality.

    Tests verify correct behavior of convolution operations applied to
    fields with various configurations, including shape validation and
    component matching.
    """

    def template_test_apply_in_2D_field(self, nb_field_components: int):
        """Test convolution operator application on 2D fields.

        This test verifies that the ConvolutionOperator correctly applies a
        stencil to a nodal field and produces output at quadrature points
        with the expected values. It tests the core functionality of the
        convolution operation.

        The convolution is defined mathematically as:
            f_{o,c,q,p} = sum_n sum_k s_{o,q,n,k} g_{c,n,p-k}
        where:
            - f: output field at quadrature points
            - g: input field at nodal points
            - s: stencil
            - o: operator index
            - c: component index
            - q: quadrature point index
            - n: nodal point index (within stencil)
            - p: pixel coordinate
            - k: stencil offset

        Parameters
        ----------
        nb_field_components : int
            Number of components in the field (1 for scalar, 3 for
            vector, etc.)

        Test Coverage
        -------
        - Correct shape of output field
        - Correct computation of convolution values
        - Proper handling of multiple components
        - Periodic boundary conditions (via np.roll)
        """
        # Parameters
        nb_x_pts = 5  # number of pixels in x axis
        nb_y_pts = 4  # number of pixels in y axis
        nb_stencil_x = 2  # number of stencil points in x axis
        nb_stencil_y = 3  # number of stencil points in y axis
        nb_operators = 3  # number of operators
        nb_quad_pts = 2  # number of quadrature points
        nb_nodal_pts = 1  # number of nodal pixel points

        # Create the operator
        conv_kern0 = np.array([[1, 0, 0], [0, 0, 0]], dtype=float)
        conv_kern1 = np.array([[0, 1, 0], [0, 0, 0]], dtype=float)
        conv_kern2 = np.array([[0, 0, 1], [0, 0, 0]], dtype=float)
        conv_kern3 = np.array([[0, 0, 0], [1, 0, 0]], dtype=float)
        conv_kern4 = np.array([[0, 0, 0], [0, 1, 0]], dtype=float)
        conv_kern5 = np.array([[0, 0, 0], [0, 0, 1]], dtype=float)
        stencil_oqij = np.array(
            [
                [conv_kern0, conv_kern3],  # Operator 1 for both quadrature points
                [conv_kern1, conv_kern4],  # Operator 2 for both quadrature points
                [conv_kern2, conv_kern5],  # Operator 3 for both quadrature points
            ]
        )
        assert stencil_oqij.shape == (
            nb_operators,
            nb_quad_pts,
            nb_stencil_x,
            nb_stencil_y,
        )
        d_op = muGrid.ConvolutionOperator([0, 0], stencil_oqij)

        # Check that convolution operator has correct shape
        assert d_op.nb_operators == nb_operators
        assert d_op.nb_quad_pts == nb_quad_pts
        assert d_op.nb_nodal_pts == nb_nodal_pts

        # Create the grid
        nb_ghosts_right = (nb_stencil_x - 1, nb_stencil_y - 1)
        fc = muGrid.GlobalFieldCollection(
            (nb_x_pts, nb_y_pts),
            sub_pts={"quad": nb_quad_pts},
            nb_ghosts_right=nb_ghosts_right,
        )

        # A nodal field with some sequence as values
        nodal = real_field(fc, "nodal-value", nb_field_components)
        nodal_vals = (1 + np.arange(nb_field_components * nb_x_pts * nb_y_pts)).reshape(
            nb_field_components, nb_x_pts, nb_y_pts
        )
        # pad to mimic a periodic boundary
        pad_width = ((0, 0), (0, nb_stencil_x - 1), (0, nb_stencil_y - 1))
        nodal.pg[...] = np.pad(nodal_vals, pad_width, "wrap")

        # Create a quadrature field to store the result
        quad = real_field(fc, "quad-grad", (nb_field_components, nb_operators), "quad")

        # Check that quadrature field has correct shape
        assert quad.s.shape == (
            nb_field_components,
            nb_operators,
            nb_quad_pts,
            nb_x_pts,
            nb_y_pts,
        )

        # Apply the gradient operator to the `nodal` field and write result to the
        # `quad` field
        d_op.apply(nodal._cpp, quad._cpp)

        # Compute the reference value
        # Create a pack of nodal values, each with a different offset
        offset_00 = nodal.p
        offset_10 = np.roll(nodal.p, (-1, 0), axis=(-2, -1))
        offset_01 = np.roll(nodal.p, (0, -1), axis=(-2, -1))
        offset_11 = np.roll(nodal.p, (-1, -1), axis=(-2, -1))
        offset_02 = np.roll(nodal.p, (0, -2), axis=(-2, -1))
        offset_12 = np.roll(nodal.p, (-1, -2), axis=(-2, -1))
        offset_nodes_ijcxy = np.array(
            [[offset_00, offset_01, offset_02], [offset_10, offset_11, offset_12]]
        )
        assert offset_nodes_ijcxy.shape == (
            nb_stencil_x,
            nb_stencil_y,
            nb_field_components,
            nb_x_pts,
            nb_y_pts,
        )

        grad_ref_ocqxy = np.einsum(
            "oqij,ijcxy->coqxy", stencil_oqij, offset_nodes_ijcxy
        )
        assert grad_ref_ocqxy.shape == (
            nb_field_components,
            nb_operators,
            nb_quad_pts,
            nb_x_pts,
            nb_y_pts,
        )

        # print(f"actual\n{quad.s}")
        # print(f"desired\n{grad_ref_ocqxy}")
        # Check
        np.testing.assert_allclose(quad.s, grad_ref_ocqxy)

    # Test cases of a scalar field & a vector field
    test_apply_2D_field_scalar = functools.partialmethod(
        template_test_apply_in_2D_field, nb_field_components=1
    )
    test_apply_2D_field_3D_vector = functools.partialmethod(
        template_test_apply_in_2D_field, nb_field_components=3
    )


@pytest.mark.skip(
    reason="Known limitation: Local fields from CartesianDecomposition "
    "cause segfault. Exception checking for is_global() exists in C++ but "
    "exception propagation needs review."
)
def test_malformed_convolution_input(comm):
    """Test convolution operator with local fields raises error.

    The ConvolutionOperator is designed to work only with global fields.
    Local fields from CartesianDecomposition should be rejected with a
    clear error message about field type.

    Setup
    -----
    - Uses CartesianDecomposition to create local fields (wrong!)
    - Attempts to apply ConvolutionOperator to these local fields

    Expected behavior
    -----------------
    Should raise a RuntimeError with message about field type mismatch.
    """
    nb_domain_grid_pts = [4, 6]

    left_ghosts = [1, 1]
    right_ghosts = [1, 1]

    decomp = muGrid.CartesianDecomposition(
        comm,
        nb_domain_grid_pts,
        suggest_subdivisions(2, comm.size),
        left_ghosts,
        right_ghosts,
    )

    fc = decomp.collection

    # Get nodal field
    nodal_field1 = fc.real_field("nodal-field1", (2,))
    nodal_field2 = fc.real_field("nodal-field2", (2,))

    impulse_locations = (nodal_field1.icoordsg[0] == 0) & (
        nodal_field1.icoordsg[1] == 0
    )
    nodal_field1.pg[0, impulse_locations] = 1
    nodal_field2.pg[1, impulse_locations] = 1

    shift = np.array(
        [
            [
                [
                    [1, 0, 0],
                    [0, 0, 0],
                    [0, 0, 0],
                ],
                [
                    [0, 0, 0],
                    [0, 0, 0],
                    [0, 0, 1],
                ],
            ],
        ],
    )
    # Stencil shape: (nb_operators=1, nb_quad_pts=2, stencil_y=3,
    # stencil_x=3) But we're using it on a local field collection
    # (after decomposition)
    shift_op = muGrid.ConvolutionOperator([-1, -1], shift)
    # This should raise an error because fields are local, not global
    with pytest.raises(RuntimeError):
        shift_op.apply(nodal_field1, nodal_field2)


def test_convolution_component_mismatch_global():
    """Test that mismatched component counts are caught.

    The ConvolutionOperator requires that:
        output_components == nb_operators * input_components

    This test verifies that the operator validates this constraint and
    raises an appropriate error rather than causing undefined behavior.

    Setup
    -----
    - Creates a 2D grid with a simple stencil (1 operator)
    - Creates an input nodal field with 2 components
    - Creates an output quadrature field with incorrect component count
      (3 instead of 2)
    - The correct count should be: 1 operator × 2 components = 2
      components

    Expected behavior
    -----------------
    Should raise a RuntimeError with a message explaining the component
    mismatch.
    """
    nb_x_pts = 3
    nb_y_pts = 3

    # Create a simple stencil with 1 operator and 1 quadrature point
    stencil = np.array(
        [
            [
                [1, 0],  # First row of stencil points
                [0, 1],
            ]
        ]
    )

    conv_op = muGrid.ConvolutionOperator([0, 0], stencil)

    # Create field collection
    fc = muGrid.GlobalFieldCollection((nb_x_pts, nb_y_pts), sub_pts={"quad": 1})

    # Input field: 2 components on nodal points (default)
    nodal_field = fc.real_field("nodal", (2,))

    # Output field with WRONG component count
    # Should be: nb_operators (1) × nb_components (2) = 2 components
    quad_field = fc.real_field("quad-wrong", (3,), "quad")

    # This should raise an error about component mismatch
    with pytest.raises(RuntimeError):
        conv_op.apply(nodal_field, quad_field)


def test_convolution_component_mismatch_multiple_operators_global():
    """Test component validation with multiple operators.

    When a ConvolutionOperator has multiple operators (e.g., for gradient
    with multiple directional derivatives), the output field must have
    components = nb_operators * nb_input_components.

    Setup
    -----
    - Creates a 2D grid with a stencil having 3 operators
    - Creates an input nodal field with 1 component
    - Creates an output quadrature field with wrong component count
    - The correct count should be: 3 operators × 1 component = 3
      components

    Expected behavior
    -----------------
    Should raise a RuntimeError with a message explaining the component
    mismatch.
    """
    nb_x_pts = 3
    nb_y_pts = 3

    # Create stencil with 3 operators
    stencil = np.zeros((3, 1, 2, 2))  # 3 ops, 1 quad pt, 2x2 stencil
    stencil[0, 0, 0, 0] = 1.0
    stencil[1, 0, 0, 1] = 1.0
    stencil[2, 0, 1, 0] = 1.0

    conv_op = muGrid.ConvolutionOperator([0, 0], stencil)

    # Create field collection
    fc = muGrid.GlobalFieldCollection((nb_x_pts, nb_y_pts), sub_pts={"quad": 1})

    # Input field: 1 component
    nodal_field = fc.real_field("nodal-scalar", (1,))

    # Output field with WRONG component count (should be 3)
    quad_field = fc.real_field("quad-wrong", (2,), "quad")

    # This should raise an error
    with pytest.raises(RuntimeError):
        conv_op.apply(nodal_field, quad_field)


def test_convolution_wrong_input_subpt_type_global():
    """Test that mismatched sub-point counts are caught.

    The ConvolutionOperator requires that input and output fields have
    the same number of sub-points per pixel. This test verifies that
    fields with mismatched sub-point counts are rejected with a clear
    error message.

    Setup
    -----
    - Creates a 2D grid with a simple stencil
    - Creates an input field with 2 sub-points (quadrature)
    - Creates an output field with 1 sub-point (nodal)
    - Tries to apply the operator

    Expected behavior
    -----------------
    Should raise a RuntimeError with message about sub-point count
    mismatch.
    """
    nb_x_pts = 3
    nb_y_pts = 3

    # Create a simple stencil
    stencil = np.array([[[1, 0], [0, 1]]])
    conv_op = muGrid.ConvolutionOperator([0, 0], stencil)

    # Create field collection with different sub-point counts
    fc_nodal = muGrid.GlobalFieldCollection((nb_x_pts, nb_y_pts))
    fc_quad = muGrid.GlobalFieldCollection((nb_x_pts, nb_y_pts), sub_pts={"quad": 2})

    # Input field with 2 sub-points (from quad collection)
    nodal_field_wrong = fc_quad.real_field("quad-field", (1,), "quad")

    # Output field with 1 sub-point (from nodal collection)
    quad_field = fc_nodal.real_field("quad-output", (1,))

    # This should raise an error about sub-point count mismatch
    with pytest.raises(RuntimeError):
        conv_op.apply(nodal_field_wrong, quad_field)


def test_convolution_wrong_output_subpt_type_global():
    """Test that mismatched output sub-point count is caught.

    The ConvolutionOperator requires that input and output fields have
    the same number of sub-points per pixel. This test verifies that
    an output field with wrong sub-point count is rejected with a clear
    error message.

    Setup
    -----
    - Creates a 2D grid with a simple stencil
    - Creates an input field with 1 sub-point (nodal)
    - Creates an output field with 2 sub-points (quadrature)
    - Tries to apply the operator

    Expected behavior
    -----------------
    Should raise a RuntimeError with message about sub-point count
    mismatch.
    """
    nb_x_pts = 3
    nb_y_pts = 3

    # Create a simple stencil
    stencil = np.array([[[1, 0], [0, 1]]])
    conv_op = muGrid.ConvolutionOperator([0, 0], stencil)

    # Create field collections with different sub-point counts
    fc_nodal = muGrid.GlobalFieldCollection((nb_x_pts, nb_y_pts))
    fc_quad = muGrid.GlobalFieldCollection((nb_x_pts, nb_y_pts), sub_pts={"quad": 2})

    # Input field with 1 sub-point (from nodal collection)
    nodal_field = fc_nodal.real_field("nodal", (1,))

    # Output field with 2 sub-points (from quad collection) - wrong!
    quad_field_wrong = fc_quad.real_field("output-quad", (1,), "quad")

    # This should raise an error about sub-point count mismatch
    with pytest.raises(RuntimeError):
        conv_op.apply(nodal_field, quad_field_wrong)


# =============================================================================
# GPU-specific convolution tests
# =============================================================================


# Use compile-time feature flag for GPU availability
GPU_AVAILABLE = muGrid.has_gpu

# Try to import CuPy for GPU tests
try:
    import cupy as cp

    HAS_CUPY = True
except ImportError:
    HAS_CUPY = False
    cp = None


class ConvolutionOperatorHostCheck(unittest.TestCase):
    """Test suite for ConvolutionOperator on host memory.

    These tests explicitly verify that host operations work correctly
    and return numpy arrays.
    """

    def test_host_convolution_returns_numpy(self):
        """Test that convolution on host fields returns numpy arrays."""
        nb_x_pts = 5
        nb_y_pts = 4
        nb_quad_pts = 2

        # Create stencil
        stencil = np.array(
            [
                [
                    [[1, 0], [0, 0]],  # Operator 0, quad pt 0
                    [[0, 1], [0, 0]],  # Operator 0, quad pt 1
                ]
            ]
        )
        conv_op = muGrid.ConvolutionOperator([0, 0], stencil)

        # Create host field collection
        fc = muGrid.GlobalFieldCollection(
            (nb_x_pts, nb_y_pts), sub_pts={"quad": nb_quad_pts}, nb_ghosts_right=(1, 1)
        )

        # Verify host collection
        self.assertFalse(fc.is_on_device)
        self.assertEqual(
            fc.memory_location, muGrid.GlobalFieldCollection.MemoryLocation.Host
        )

        # Create fields
        nodal = real_field(fc, "nodal", (2,))
        quad = real_field(fc, "quad", (2, 1), "quad")

        # Verify fields are on host
        self.assertFalse(nodal.is_on_gpu)
        self.assertFalse(quad.is_on_gpu)
        self.assertEqual(nodal.device, "cpu")
        self.assertEqual(quad.device, "cpu")

        # Verify arrays are numpy
        self.assertIsInstance(nodal.s, np.ndarray)
        self.assertIsInstance(quad.s, np.ndarray)

        # Initialize and apply
        nodal.pg[...] = np.random.rand(*nodal.pg.shape)
        conv_op.apply(nodal._cpp, quad._cpp)

        # Result should still be numpy
        self.assertIsInstance(quad.s, np.ndarray)

    def test_host_convolution_correctness(self):
        """Test that host convolution produces correct results."""
        nb_x_pts = 4
        nb_y_pts = 4

        # Simple identity stencil
        stencil = np.array([[[[1, 0], [0, 0]]]])
        conv_op = muGrid.ConvolutionOperator([0, 0], stencil)

        fc = muGrid.GlobalFieldCollection(
            (nb_x_pts, nb_y_pts), sub_pts={"quad": 1}, nb_ghosts_right=(1, 1)
        )

        nodal = real_field(fc, "nodal", (1,))
        quad = real_field(fc, "quad", (1,), "quad")

        # Set input to known values
        test_vals = np.arange(nb_x_pts * nb_y_pts, dtype=float).reshape(
            1, nb_x_pts, nb_y_pts
        )
        nodal.pg[...] = np.pad(test_vals, ((0, 0), (0, 1), (0, 1)), "wrap")

        conv_op.apply(nodal._cpp, quad._cpp)

        # With identity stencil, output should match input
        np.testing.assert_allclose(
            quad.s.reshape(nb_x_pts, nb_y_pts), test_vals.reshape(nb_x_pts, nb_y_pts)
        )


@unittest.skipUnless(GPU_AVAILABLE, "GPU backend not available")
class ConvolutionOperatorDeviceCheck(unittest.TestCase):
    """Test suite for ConvolutionOperator on device (GPU) memory.

    These tests verify that convolution operations work correctly
    on GPU fields.
    """

    def setUp(self):
        self.nb_x_pts = 5
        self.nb_y_pts = 4
        self.nb_quad_pts = 2

    def test_device_collection_properties(self):
        """Test that device collection has correct properties."""
        fc = muGrid.GlobalFieldCollection(
            (self.nb_x_pts, self.nb_y_pts),
            memory_location=muGrid.GlobalFieldCollection.MemoryLocation.Device,
        )

        self.assertTrue(fc.is_on_device)
        self.assertEqual(
            fc.memory_location, muGrid.GlobalFieldCollection.MemoryLocation.Device
        )

    def test_device_field_properties(self):
        """Test that device fields have correct properties."""
        fc = muGrid.GlobalFieldCollection(
            (self.nb_x_pts, self.nb_y_pts),
            memory_location=muGrid.GlobalFieldCollection.MemoryLocation.Device,
        )

        field = real_field(fc, "test", (3,))

        self.assertTrue(field.is_on_gpu)
        device = field.device
        self.assertTrue(device.startswith("cuda:") or device.startswith("rocm:"))

    def test_device_convolution_field_creation(self):
        """Test creating fields for convolution on device."""
        # Use the constructor that accepts nb_domain_grid_pts (tuple) as first arg
        # to enable nb_ghosts_right parameter
        fc = muGrid.GlobalFieldCollection(
            (self.nb_x_pts, self.nb_y_pts),
            sub_pts={"quad": self.nb_quad_pts},
            nb_ghosts_right=(1, 1),
            memory_location=muGrid.GlobalFieldCollection.MemoryLocation.Device,
        )

        nodal = real_field(fc, "nodal", (2,))
        quad = real_field(fc, "quad", (2, 1), "quad")

        # Both fields should be on device
        self.assertTrue(nodal.is_on_gpu)
        self.assertTrue(quad.is_on_gpu)

    def test_device_convolution_apply(self):
        """Test that convolution can be applied on device fields."""
        stencil = np.array(
            [
                [
                    [[1, 0], [0, 0]],
                    [[0, 1], [0, 0]],
                ]
            ]
        )
        conv_op = muGrid.ConvolutionOperator([0, 0], stencil)

        # Use the constructor that accepts nb_domain_grid_pts (tuple) as first arg
        # to enable nb_ghosts_right parameter
        fc = muGrid.GlobalFieldCollection(
            (self.nb_x_pts, self.nb_y_pts),
            sub_pts={"quad": self.nb_quad_pts},
            nb_ghosts_right=(1, 1),
            memory_location=muGrid.GlobalFieldCollection.MemoryLocation.Device,
        )

        nodal = real_field(fc, "nodal", (2,))
        quad = real_field(fc, "quad", (2, 1), "quad")

        # Initialize (set_zero should work on device)
        nodal._cpp.set_zero()
        quad._cpp.set_zero()

        # Apply should not raise
        conv_op.apply(nodal._cpp, quad._cpp)


@unittest.skipUnless(GPU_AVAILABLE and HAS_CUPY, "GPU backend or CuPy not available")
class ConvolutionOperatorCuPyCheck(unittest.TestCase):
    """Test suite for ConvolutionOperator with CuPy arrays.

    These tests verify that GPU convolution operations return CuPy arrays
    and produce correct results that can be verified against numpy.
    """

    def setUp(self):
        self.nb_x_pts = 5
        self.nb_y_pts = 4
        self.nb_quad_pts = 2

    def test_device_convolution_returns_cupy(self):
        """Test that convolution on device fields returns CuPy arrays."""
        stencil = np.array(
            [
                [
                    [[1, 0], [0, 0]],
                    [[0, 1], [0, 0]],
                ]
            ]
        )
        conv_op = muGrid.ConvolutionOperator([0, 0], stencil)

        # Use the constructor that accepts nb_domain_grid_pts (tuple) as first arg
        # to enable nb_ghosts_right parameter
        fc = muGrid.GlobalFieldCollection(
            (self.nb_x_pts, self.nb_y_pts),
            sub_pts={"quad": self.nb_quad_pts},
            nb_ghosts_right=(1, 1),
            memory_location=muGrid.GlobalFieldCollection.MemoryLocation.Device,
        )

        nodal = real_field(fc, "nodal", (2,))
        quad = real_field(fc, "quad", (2, 1), "quad")

        # Arrays should be CuPy
        self.assertIsInstance(nodal.s, cp.ndarray)
        self.assertIsInstance(quad.s, cp.ndarray)

        # Apply convolution
        conv_op.apply(nodal._cpp, quad._cpp)

        # Result should still be CuPy
        self.assertIsInstance(quad.s, cp.ndarray)

    def test_device_convolution_correctness(self):
        """Test that device convolution produces correct results.

        Compare GPU convolution results against CPU reference.
        """
        nb_operators = 2
        stencil = np.array(
            [
                [
                    [[1, 0], [0, 0]],
                    [[0, 1], [0, 0]],
                ],
                [
                    [[0, 0], [1, 0]],
                    [[0, 0], [0, 1]],
                ],
            ]
        )
        conv_op = muGrid.ConvolutionOperator([0, 0], stencil)

        # Create host reference
        fc_host = muGrid.GlobalFieldCollection(
            (self.nb_x_pts, self.nb_y_pts),
            sub_pts={"quad": self.nb_quad_pts},
            nb_ghosts_right=(1, 1),
        )
        nodal_host = real_field(fc_host, "nodal", (2,))
        quad_host = real_field(fc_host, "quad", (2, nb_operators), "quad")

        # Create device fields
        # Use the constructor that accepts nb_domain_grid_pts (tuple) as first arg
        # to enable nb_ghosts_right parameter
        fc_device = muGrid.GlobalFieldCollection(
            (self.nb_x_pts, self.nb_y_pts),
            sub_pts={"quad": self.nb_quad_pts},
            nb_ghosts_right=(1, 1),
            memory_location=muGrid.GlobalFieldCollection.MemoryLocation.Device,
        )
        nodal_device = real_field(fc_device, "nodal", (2,))
        quad_device = real_field(fc_device, "quad", (2, nb_operators), "quad")

        # Initialize with same random data
        test_data = np.random.rand(*nodal_host.pg.shape)
        nodal_host.pg[...] = test_data
        nodal_device.pg[...] = cp.asarray(test_data)

        # Apply on both
        conv_op.apply(nodal_host._cpp, quad_host._cpp)
        conv_op.apply(nodal_device._cpp, quad_device._cpp)

        # Compare results
        host_result = quad_host.s
        device_result = cp.asnumpy(quad_device.s)

        np.testing.assert_allclose(device_result, host_result, rtol=1e-10)

    def test_host_device_mixed_workflow(self):
        """Test workflow transferring data between host and device."""
        stencil = np.array([[[[1, 0], [0, 0]]]])
        conv_op = muGrid.ConvolutionOperator([0, 0], stencil)

        # Create host collection for initial data
        fc_host = muGrid.GlobalFieldCollection(
            (self.nb_x_pts, self.nb_y_pts), sub_pts={"quad": 1}, nb_ghosts_right=(1, 1)
        )
        nodal_host = real_field(fc_host, "nodal", (1,))

        # Create device collection for computation
        # Use the constructor that accepts nb_domain_grid_pts (tuple) as first arg
        # to enable nb_ghosts_right parameter
        fc_device = muGrid.GlobalFieldCollection(
            (self.nb_x_pts, self.nb_y_pts),
            sub_pts={"quad": 1},
            nb_ghosts_right=(1, 1),
            memory_location=muGrid.GlobalFieldCollection.MemoryLocation.Device,
        )
        nodal_device = real_field(fc_device, "nodal", (1,))
        quad_device = real_field(fc_device, "quad", (1,), "quad")

        # Initialize host data
        test_vals = np.random.rand(*nodal_host.pg.shape)
        nodal_host.pg[...] = test_vals

        # Transfer to device
        nodal_device.pg[...] = cp.asarray(nodal_host.pg)

        # Compute on device
        conv_op.apply(nodal_device._cpp, quad_device._cpp)

        # Transfer result back to host
        result_on_host = cp.asnumpy(quad_device.s)

        # Verify result is numpy array
        self.assertIsInstance(result_on_host, np.ndarray)



import numpy as np
import muGrid

# Grid parameters
nx, ny = 64, 64
Lx, Ly = 2 * np.pi, 2 * np.pi
dx, dy = Lx / nx, Ly / ny

engine = muGrid.FFTEngine([nx, ny])

# Create fields
f = engine.real_space_field("f")
f_hat = engine.fourier_space_field("f_hat")
grad_x = engine.real_space_field("grad_x")
grad_y = engine.real_space_field("grad_y")
grad_hat = engine.fourier_space_field("grad_hat")

# Initialize with a smooth function
x = np.linspace(0, Lx, nx, endpoint=False)
y = np.linspace(0, Ly, ny, endpoint=False)
X, Y = np.meshgrid(x, y, indexing='ij')
f.s[0, 0, :, :] = np.sin(X) * np.cos(Y)

# Forward FFT
engine.fft(f, f_hat)

# Compute wavevectors
kx = 2 * np.pi * np.array(muGrid.rfft_freqind(nx)) / Lx
ky = 2 * np.pi * np.array(muGrid.fft_freqind(ny)) / Ly
KX, KY = np.meshgrid(kx, ky, indexing='ij')

# Derivative in x: multiply by i*kx
grad_hat.s[0, 0, :, :] = 1j * KX * f_hat.s[0, 0, :, :]
engine.ifft(grad_hat, grad_x)
grad_x.s[:] *= engine.normalisation

# Derivative in y: multiply by i*ky
grad_hat.s[0, 0, :, :] = 1j * KY * f_hat.s[0, 0, :, :]
engine.ifft(grad_hat, grad_y)
grad_y.s[:] *= engine.normalisation

# Verify: d/dx[sin(x)cos(y)] = cos(x)cos(y)
expected_grad_x = np.cos(X) * np.cos(Y)
np.testing.assert_allclose(grad_x.s[0, 0], expected_grad_x, atol=1e-10)



"""
Collection of simple parallel solvers
"""

from muGrid import real_field

# Import the C++ extension module
# Try relative import first (for installed wheels),
# fall back to absolute (for development)
try:
    from ._muGrid import Communicator, FieldCollection
except ImportError:
    from _muGrid import Communicator, FieldCollection

from .Field import wrap_field


def conjugate_gradients(
    comm: Communicator,
    fc: FieldCollection,
    hessp: callable,
    b,
    x,
    tol: float = 1e-6,
    maxiter: int = 1000,
    callback: callable = None,
):
    """
    Conjugate gradient method for matrix-free solution of the linear problem
    Ax = b, where A is represented by the function hessp (which computes the
    product of A with a vector). The method iteratively refines the solution x
    until the residual ||Ax - b|| is less than tol or until maxiter iterations
    are reached.

    Parameters
    ----------
    comm : muGrid.Communicator
        Communicator for parallel processing.
    fc : muGrid.FieldCollection
        Collection holding temporary fields of the CG algorithm.
    hessp : callable
        Function that computes the product of the Hessian matrix A with a vector.
    b : muGrid.Field
        Right-hand side vector (C++ field or wrapped Field).
    x : muGrid.Field
        Initial guess for the solution (C++ field or wrapped Field).
    tol : float, optional
        Tolerance for convergence. The default is 1e-6.
    maxiter : int, optional
        Maximum number of iterations. The default is 1000.
    callback : callable, optional
        Function to call after each iteration with the current solution, residual,
        and search direction.

    Returns
    -------
    x : array_like
        Approximate solution to the system Ax = b. (Same as input field x.)
    """
    tol_sq = tol * tol

    # Wrap fields for array access if they are not already wrapped
    b = wrap_field(b)
    x = wrap_field(x)

    p = real_field(fc, "cg-search-direction")
    Ap = real_field(fc, "cg-hessian-product")

    hessp(x, Ap)
    p.s[...] = b.s - Ap.s
    r = p.s.copy()  # residual

    if callback:
        callback(0, x.s, r, p.s)

    rr = comm.sum(r.ravel().dot(r.ravel()))  # initial residual dot product
    if rr < tol_sq:
        return x

    for iteration in range(maxiter):
        # Compute Hessian product
        hessp(p, Ap)

        # Update x (and residual)
        pAp = comm.sum(p.s.ravel().dot(Ap.s.ravel()))
        if pAp <= 0:
            raise RuntimeError("Hessian is not positive definite")

        alpha = rr / pAp
        x.s[...] += alpha * p.s
        r -= alpha * Ap.s

        if callback:
            callback(iteration + 1, x.s, r, p.s)

        # Check convergence
        next_rr = comm.sum(r.ravel().dot(r.ravel()))
        if next_rr < tol_sq:
            return x

        # Update search direction
        beta = next_rr / rr
        rr = next_rr
        p.s[...] *= beta
        p.s[...] += r

    raise RuntimeError("Conjugate gradient algorithm did not converge")
